#coding:utf-8

import pandas as pd
import numpy as np
import json
import os
from scipy.stats.mstats import gmean
import random
from joblib import Parallel, delayed

import mpbn
import mpbn.simulation as mpbn_sim
from tqdm import tqdm
import gc

from NORDic.UTILS.utils_state import compare_states

####################
## Spread process ##
####################

def compute_similarities(f, x0, A, A_WT, gene_outputs, nb_sims, experiments, repeat=1, exp_name="", quiet=False):
    '''
        Compute similarities between any attractor in WT and in mutants, weighted by their probabilities
        @param\tf\tBoolean Network (MPBN) object: the mutated network
        @param\tx0\tMPBN object: initial state
        @param\tA\tAttractor list: list of attractors in mutant
        @param\tA_WT\tAttractor list: list of attractors in WT
        @param\tgene_outputs\tPython character string list: list of node names to check
        @param\tnb_sims\tPython integer: number of iterations to compute the probabilities
        @param\texperiments\tPython dictionary list: list of experiments (different rates/depths)
        @param\trepeat\tPython integer[default=1]: how many times should these experiments be repeated
        @param\texp_name\tPython character string[default=""]: printed info about the experiment (if quiet=True)
        @param\tquiet\tPython bool[default=False]
        @return\tsim\tPython float: change in attractors induced by the mutation
    '''
    for exp in experiments:
        if "name" not in exp:
            continue
        dists = []
        for _ in range(repeat):
            rates = getattr(mpbn_sim, f"{exp['rates']}_rates")
            depth = getattr(mpbn_sim, f"{exp['depth']}_depth")
            rates_args = exp.get("rate_args", {})
            depth_args = exp.get("depth_args", {})
            if (not quiet):
                print(exp_name+" "*int(len(exp_name)>0)+(f"- {depth.__name__}{depth_args}\t{rates.__name__}{rates_args}"))
            probs = mpbn_sim.estimate_reachable_attractors_probabilities(f, x0, A, nb_sims, depth(f, **depth_args), rates(f, **rates_args))
            attrs = pd.DataFrame({"MUT_%d"%ia: a for ia, a in enumerate(A)}).replace("*",np.nan).astype(float)
            probs = {i: x for i,x in list(probs.items()) if (x>0)}
            attrs = attrs[[attrs.columns[i] for i in list(probs.keys())]]
            ## if too many attractors, select the most common ones
            if (attrs.shape[1]>45000):
                idx_common = np.argsort([probs[i] for i in range(len(probs))]).tolist()
                idx_common.reverse()
                idx_common = idx_common[:45000]
                attrs = attrs[[attrs.columns[i] for i in idx_common]]
                probs = {i:probs[i] for i in idx_common}
                sprobs = sum(list(probs.values()))
                probs = {i:probs[i]*100/sprobs for i in probs}
            probs = np.array([probs[ia]/100 for ia in probs])
            attrs_init = pd.DataFrame({"WT_%d"%ia: a for ia, a in enumerate(A_WT)}).replace("*",np.nan).astype(float)
            sims, nb_gene = compare_states(attrs, attrs_init, gene_outputs)
            assert nb_gene == len(gene_outputs)
            sims = probs.T.dot(sims)
            dt = 1-np.max(sims) #max: minimum of change in (*different*) attractors induced by the subset S
            dists.append(dt)
    return np.mean(dists) if (repeat>1) else dists[0]

def run_experiments(network_name, spreader, gene_list, state, gene_outputs, simu_params, quiet=False):
    ## Create file
    import json
    from subprocess import call as sbcall
    perts, perts_S = [(state.loc[[g for g in lst if (g in state.index)]]+1)%2 for lst in [gene_list, spreader]]
    experiments, nb_sims = [{"name": "mpsim", "rates": simu_params.get("rates", "fully_asynchronous"), "depth": simu_params.get("depth", "constant_unitary")}], simu_params["nb_sims"]
    experiments_di = {
        "bnet_file": network_name,
        "init_active": list(state.loc[state[state.columns[0]]==1].index),
        "nb_sims": nb_sims,
        "mutants": {g+"_"+("KO" if (perts.loc[g][perts.columns[0]]==0) else "OE")+" {"+str(spreader)+"}": {gx: str(pd.concat((perts,perts_S),axis=0).loc[gx][perts.columns[0]]) for gx in [g]+spreader} for g in list(perts.index)},
        "experiments": experiments,
    }
    with open("experiments.json", "w") as f:
        json.dump(experiments_di)
    sbcall("mpbn_sim --save experiments.json", shell=True)

def spread(network_name, spreader, gene_list, state, gene_outputs, simu_params, seednb=0, quiet=False):
    '''
        Compute the spread of each gene in @gene_inputs+@spreader with initial state @state on genes @gene_outputs
        Here, the (single state) spread is defined as the indicator of the emptyness of the intersection between WT and mutant attractors
        @param\tnetwork_name\tPython character string: filename of the network in .bnet (needs to be pickable)
        @param\tspreader\tPython character string list: subset of node names
        @param\tgene_list\tPython character string list: list of node names to perturb in addition to the spreader
        @param\tstate\tPandas DataFrame: binary initial state rows/[genes] x columns/[values in {-1,0,1}]
        @param\tgene_outputs\tPython character string list: list of node names to check
        @param\tsimu_params\tPython dictionary: arguments to MPBN-SIM
        @param\tseednb\tPython integer[default=0]
        @param\tquiet\tPython bool[default=False]
        @return\tspds\tPython float dictionary: change in mutant attractor states for each gene in @gene_list
        that is, the similarity between any attractor reachable from @state in WT and any in mutant spreader+{g} where g in gene_list
    '''
    random.seed(seednb)
    np.random.seed(seednb)
    ## 1. Load the Boolean network
    f = mpbn.load(network_name)
    ## 2. Create the initial profile
    if (not quiet):
        print("\t<NORD_PMR> Initial state %s (gene(s):%s)" % (state.columns[0], gene_list[0]))
    x0 = f.zero()
    for i in list(state.loc[state[state.columns[0]]==1].index):
        x0[i] = 1
    ## 3. Get the reachable attractors from initial state in the absence of perturbation ("wild type")
    experiments, nb_sims = [{"name": "mpsim", "rates": simu_params.get("rates", "fully_asynchronous"), "depth": simu_params.get("depth", "constant_unitary")}], simu_params["nb_sims"]
    A_WT = [a for a in tqdm(list(f.attractors(reachable_from=x0)))]
    exp = experiments[0]
    rates = getattr(mpbn_sim, f"{exp['rates']}_rates")
    depth = getattr(mpbn_sim, f"{exp['depth']}_depth")
    rates_args = exp.get("rate_args", {})
    depth_args = exp.get("depth_args", {})
    probs_WT = mpbn_sim.estimate_reachable_attractors_probabilities(f, x0, A_WT, nb_sims, depth(f, **depth_args), rates(f, **rates_args))
    probs_WT = {i: x for i,x in list(probs_WT.items()) if (x>0)}
    A_WT = [A_WT[i] for i in list(probs_WT.keys())]
    if (not quiet):
        print("%d wild type attractors with proba > 0 (initial state %s)" % (len(A_WT), state.columns[0]))
    ## if too many attractors, select the most common ones
    if (len(A_WT)>45000):
        idx_common_WT = np.argsort([probs_WT[i] for i in range(len(probs_WT))]).tolist()
        idx_common_WT.reverse()
        idx_common_WT = idx_common_WT[:45000]
        A_WT = [A_WT[i] for i in idx_common_WT]
        if (not quiet):
            print("> reduced to %d wild type attractors (initial state %s, %d perc. of all attractors)" % (len(A_WT), np.sum([probs_WT[i] for i in idx_common_WT])))
    ## 4. Create the mutated networks
    def patch_model(f, patch):
        f = mpbn.MPBooleanNetwork(f)
        for i, fi in patch.items():
            f[i] = fi
        return f
    perts, perts_S = [(state.loc[[g for g in lst if (g in state.index)]]+1)%2 for lst in [gene_list, spreader]]
    mutants = {g: {gx: str(pd.concat((perts,perts_S),axis=0).loc[gx][perts.columns[0]]) for gx in [g]+spreader} for g in list(perts.index)}
    f_mutants = {name: patch_model(f, patch) for name, patch in mutants.items()}
    ## 5. Get the reachable attractors from initial state in the presence of mutations ("mutants" KO/OE)
    ## 6. Estimate probabilities of attractors from Mutants and compute similarities
    spds = [0 if (name_g not in f_mutants) else compute_similarities(f_mutants[name_g], x0, [a for a in tqdm(list(f_mutants[name_g].attractors(reachable_from=x0)))], A_WT, [g for g in gene_outputs if (g not in [name_g]+spreader)], nb_sims, experiments, exp_name="Gene %s (%d/%d) in state %s" % (name_g, ig+1, len(gene_list), state.columns[0]), quiet=quiet) for ig, name_g in enumerate(gene_list)]
    return spds

def spread_multistate(network_name, spreader, gene_list, states, gene_outputs, im_params, simu_params, quiet=False):
    '''
        Compute the spread of each gene in @gene_inputs+@spreader with initial states in @states on genes @gene_outputs
        Here, the (single state) spread is defined as the indicator of the emptyness of the intersection between WT and mutant attractors
        @param\tnetwork_name\tPython character string: filename of the network in .bnet (needs to be pickable)
        @param\tspreader\tPython character string list: subset of node names
        @param\tgene_list\tPython character string list: list of node names to perturb in addition to the spreader
        @param\tstates\tPandas DataFrame: binary initial state rows/[genes] x columns/[state ID]
        @param\tgene_outputs\tPython character string list: list of node names to check
        @param\tim_params\tPython dictionary: arguments to Influence Maximization
        @param\tsimu_params\tPython dictionary: arguments to MPBN-SIM
        @param\tquiet\tPython bool[default=False]
        @return\tspds\tPython float dictionary: change in mutant attractor states for each gene in @gene_list
        that is, the geometric mean of similarities between any attractor reachable from state in @states in WT and any in mutant spreader+{g} where g in gene_list
    '''
    from multiprocessing import cpu_count
    assert simu_params.get('thread_count', 1)>=1 and simu_params.get('thread_count', 1)<=max(1,cpu_count()-2)
    if (simu_params.get('thread_count', 1)==1):
        sprds_multistate = [spread(network_name, spreader, gene_list, states[[col]], gene_outputs, simu_params, seednb=im_params.get("seed", 0)) for col in states.columns]
    else:
        if (states.shape[1]>1):
            sprds_multistate = Parallel(n_jobs=simu_params['thread_count'], backend='loky')(delayed(spread)(network_name, spreader, gene_list, states[[col]], gene_outputs, simu_params, seednb=im_params.get("seed", 0)) for col in states.columns)
        else:
            sprds_multistate = Parallel(n_jobs=simu_params['thread_count'], backend='loky')(delayed(spread)(network_name, spreader, [gene], states[[col]], gene_outputs, simu_params, seednb=im_params.get("seed", 0)) for col in states.columns for gene in gene_list)
    ## Aggregate the values across the set of initial states
    ## Genes which are not measured in the states are assigned value 0
    if (states.shape[1]>1):
        spds = [(gmean([(s[ig]+1) for s in sprds_multistate])-1) for ig, g in enumerate(gene_list)]
    else:
        spds = [sprds_multistate[ig][0] for ig, g in enumerate(gene_list)]
    return spds

#######################################
## INFLUENCE MAXIMIZATION ALGORITHM  ##
#######################################

def greedy(network_name, k, states, im_params, simu_params, save_folder=None, quiet=False):
    '''
        Greedy Influence Maximization Algorithm [Kempe et al., 2003]
        Finds iteratively the maximum spreader and adds it to the list until the list is of size k
        @param\tnetwork_name\tPython character string: bnet network
        @param\tk\tPython integer: maximum size of the spreader
        @param\tim_params\tPython dictionary or None[default=None]: parameters of the influence maximization
        @param\tstates\tPandas DataFrame or None[default=None]: list of initial states to consider
        @param\tsave_folder\tPython character string[default=None]: where to save intermediary results (if None: do not save intermediary results)
        @param\tquiet\tPython bool[default=False]
        @return\tS, spreads\tPython character string list: nodes in the spreader set, Python dictionary: spread value associated with every tested subset of nodes
    '''
    random.seed(im_params.get("seed", 0))
    np.random.seed(im_params.get("seed", 0))
    if (im_params.get("gene_inputs", None) is None):
        with open(solution_fname, "r") as f:
            network = str(f.read())
        genes = [x.split(", ")[0] for x in network.split("\n")[:-1]]
        gene_inputs = genes
    else:
        gene_inputs = im_params["gene_inputs"]
    if (im_params.get("gene_outputs", None) is None):
        with open(solution_fname, "r") as f:
            network = str(f.read())
        gene_outputs = [x.split(", ")[0] for x in network.split("\n")[:-1] if (x.split(", ")[1] not in [x.split(", ")[0], "0", "1"])]
    else:
        gene_outputs = im_params["gene_outputs"]
    S, S_unfold, spreads, start_k = [], [], {}, 0
    if (save_folder is not None and os.path.exists(save_folder+"application_regulators.json")):
        with open(save_folder+"application_regulators.json", "r") as f:
            res = json.loads(f.read())
        if (len(res)>0):
            spreads, S, S_unfold, start_k = res["spreads"], res["S"], res["S_unfold"], res["k"]
    S_spread = -1 if (len(spreads)==0 and len(S)==0) else np.max(list(spreads[str(S)].values()))
    for current_k in range(start_k+1, k+1):
        if (not quiet):
            print("<NORD_PMR> Iteration k=%d" % (current_k))
        gene_list = [g for g in gene_inputs if (g not in S_unfold)]
        ## List of Python floats in the order gene_list
        sprds_lst = spread_multistate(network_name, S_unfold, gene_list, states, gene_outputs, im_params, simu_params, quiet=quiet)
        ## can't find spreader that strictly increases the spread of the previous spreader
        sprds_arr = np.array(sprds_lst)
        if (np.max(sprds_arr)<=S_spread):
            if (not quiet):
                print("\n<NORD_PMR> /!\ Can't find a more performant spreader for k=%d" % current_k)
            break
        ## Keep ex-aequo
        ex_aequo = np.argwhere(sprds_arr==np.max(sprds_arr)).flatten().tolist()
        S_spread, nodes = np.max(sprds_arr), [gene_list[x] for x in ex_aequo]
        if (len(nodes)>1):
            nodes = tuple(nodes)
            S_unfold += nodes
            S += [nodes]
        else:
            nodes = nodes[0]
            S_unfold += [nodes]
            S += [[nodes]]
        spreads.update({str(S):{g: sprds_lst[ig] if (g not in S_unfold) else S_spread for ig, g in enumerate(gene_list)}})
        with open(save_folder+"application_regulators.json", "w") as f:
            json.dump({"spreads": spreads, "S": S, "S_unfold": S_unfold, "k": current_k}, f)
            pd.DataFrame(spreads).to_csv(save_folder+"application_regulators.csv")
    spreads_df = pd.DataFrame(spreads)
    return S, spreads_df



#coding: utf-8

from subprocess import check_output as sbcheck_output
from subprocess import call as sbcall
import json
from time import sleep
import os
import pandas as pd
import numpy as np
import cmapPy.pandasGEXpress.parse_gctx as parse_gctx
from functools import reduce

def get_user_key(fname):
    '''
        Retrieves user key for interacting with LINCS L1000 CLUE API
        @param\tfname\tPython character string: path to file containing credentials for LINCS L1000 (first line: username, second line: password, third line: user key)
        @return\tuser_key\tPython character string: identifier for the LINCS L1000 CLUE API
    '''
    with open(fname, "r") as f:
        user_key = f.read().split("\n")[2]
    return user_key

def convert_ctrlgenes_EntrezGene(taxon_id):
    '''
        Retrieves EntrezID from control genes in LINCS L1000 [1]
        @param\ttaxon_id\tPython integer: NCBI taxonomy ID
        @return\tlincs_specific_ctl_genes\tPython character string list: list of EntrezGene IDs for all genes in input list
        [1] doi.org/10.1002/psp4.12107 
    '''
    lincs_specific_ctl_genes = ["ACTB", "CHMP2A", "EEF1A1", "EMC7", "GAPDH", "GPI", "PSMB2", "PSMB4", "RAB7A", "REEP5", "SNRPD3", "TUBA1A", "VCP"]
    from NORDic.UTILS.utils_data import request_biodbnet
    lincs_specific_ctl_genes_ = [np.min(list(map(int, s.split("; ")))) for s in request_biodbnet(lincs_specific_ctl_genes, from_="Gene Symbol and Synonyms", to_="Gene ID", taxon_id=taxon_id, chunksize=100)["Gene ID"]]
    lincs_specific_ctl_genes = {lincs_specific_ctl_genes[ix]: int(x) for ix, x in enumerate(lincs_specific_ctl_genes_)}
    return lincs_specific_ctl_genes

####################
## API CALLS      ##
####################

lincs_api_url = "https://api.clue.io/api"

def build_url(endpoint, method, params, user_key=None):
    '''
        Builds the request to CLUE API
        @param\tendpoint\tPython character string: in ["sigs", "cells", "genes", "perts", "plates", "profiles", "rep_drugs", "rep_drug_indications", "pcls"]
        @param\tmethod\tPython character string: in ["count", "filter", "distinct"]
        @param\tparams\tPython dictionary: additional arguments for the request
        @param\tuser_key\tPython character string
        @return\turl\tPython character string: URL of request
    '''
    assert user_key
    assert endpoint in ["sigs", "cells", "genes", "perts", "plates", "profiles", "rep_drugs", "rep_drug_indications", "pcls"]
    assert method in ["count", "filter", "distinct"]
    request_url = "/".join([lincs_api_url, endpoint])
    convert_dict = lambda lst : '"'.join(str(lst).split("\'"))
    if (len(params) > 0):
        params_concat = "&".join([k+"="+convert_dict(params[k]) for k in list(params.keys())])
    else:
        params_concat = ""
    if (method == "count"):
        if (len(params) > 0):
            params_concat = "&".join([k+"="+convert_dict(params[k]) for k in list(params.keys())])
        else:
            params_concat = ""
        request_url += "/"+method+"?"+params_concat
    elif (method == "filter"):
        if (len(params) > 0):
            params_concat = "{"+(",".join(['"'+k+'":'+convert_dict(params[k]) for k in list(params.keys())]))+"}"
        else:
            params_concat = ""
        request_url += "?"+method+"="+params_concat
    elif (method == "distinct"):
        assert params["field"]
        if ("where" in list(params.keys())):
            request_url += "/"+method+"?where="+convert_dict(params["where"])+"&field="+params["field"]
        else:
            request_url += "/"+method+"?field="+params["field"]
    else:
        raise ValueError("<LINCS> Request method could not be identified.")
    return request_url+"&user_key="+user_key

def post_request(url, quiet=True, pause_time=1):
    '''
        Post request to API
        @param\turl\tPython character string: URL formatted as in build_url
        @param\tquiet\tPython bool[default=True]
        @param\tpause_time\tPython integer[default=1]: minimum time in seconds between each request
        @return\tdata\tPython dictionary (JSON) / Python character string list [if request was method="distinct"]
    '''
    if (not quiet):
        print("> POST "+url.split("&")[0])
    response = sbcheck_output("wget -O - -q \'" + url + "\'", shell=True)
    data = json.loads(response)
    if ("/distinct?field=" in url):
        data = list(set([y for x in data for y in x]))
    sleep(pause_time)
    return data

##<><><><><><><><><><><><><><><><><><><><><><><><><><><><>##
##   Binarization of Level 3 signatures                   ##
##<><><><><><><><><><><><><><><><><><><><><><><><><><><><>##

def binarize_via_CD(df, samples, binarize=1, nperm=10000, quiet=False):
    '''
        Run a differential expression analysis on a dataframe using Characteristic Direction (CD) [1] (implementation: www.maayanlab.net/CD/)
        @param\tdf\tPandas DataFrame: 1 transcriptional profile per column (/!\ if #genes>25,000, then the 25,000 genes with highest variance will be considered)
        @param\tsamples\tPython integer list: indicates which columns correspond to control (=1) / treated (=2) samples
        @param\tbinarize\tPython integer[default=1]: whether to return a binary signature or a real-valued column ~magnitude of change in expression
        @param\tnperm\tPython integer[default=10000]: number of iterations to build the null distribution on which p-values will be computed
        @param\tquiet\tPython bool[default=False]
        @return\tsignature\tPandas DataFrame: rows/[gene index] x columns/["aggregated"]: 0=down-regulated (DR), 1=up-regulated (UR) (if binarize=1) else <0=DR, >0=UR
        [1] doi.org/10.1186/1471-2105-15-79
    '''
    sbcheck_output("pip install git+https://github.com/Maayanlab/geode.git", shell=True)
    from geode import chdir
    assert all([s in [1,2] for s in samples])
    assert binarize in [0,1]
    assert nperm>1
    assert len(df.columns) >= 4
    assert len(df.columns) == len(samples)
    assert len(np.argwhere(np.array(samples) == 1).tolist()) > 1
    assert len(np.argwhere(np.array(samples) == 2).tolist()) > 1
    df = df.dropna()
    ## To decrease computational cost (in terms of memory)
    if (df.shape[0] > 25000):
        selected_genes = list(df.var(axis=0).sort_values(ascending=False).index)[:25000]
        if (not quiet):
            print("/!\ The number of genes has been trimmed down to the 25,000 genes with highest variance (down from %d). Consider filtering your genes beforehand!" % df.shape[0])
        df = df.loc[selected_genes]
    # @param mat NumPy matrix
    # @param samples Python integer list
    # @param calculate_sig Python integer: computes p-values if set to 1
    # @param nnull Python integer
    # @param sig_only Python integer: returns only significant genes if set to 1
    # @returns chdir_res: list of tuples (magnitude, gene, *) sorted by decreasing *absolute* magnitude
    chdir_res = chdir(df.values, samples, list(df.index), calculate_sig=binarize, nnull=nperm, sig_only=binarize)
    sign_genes = {x[1] : x[0] for x in chdir_res}
    signature = pd.DataFrame([[sign_genes.get(g, np.nan)] for g in df.index], index=df.index, columns=["aggregated"])
    if (binarize>0):
        signature = (signature > 0).astype(int)-(signature < 0).astype(int)
        signature[signature==0] = np.nan
        signature[signature<0] = 0
    return signature

####################
## DOWNLOAD FILES ##
####################

def download_file(path_to_lincs, file_name, base_url, file_sha, check_SHA=True, quiet=False):
    '''
        Downloads automatically LINCS L1000-related files from Gene Expression Omnibus (GEO) (/!\ can be time-consuming: expect waiting times up to 20 min with a good Internet connection)
        @param\tpath_to_lincs\tPython character string: path to local LINCS L1000 folder in which the files will be downloaded
        @param\tfile_name\tPython character string: file name to download on GEO
        @param\tbase_url\tPython character string: path to GEO repository
        @param\tfile_sha\tPython character string: file name of corresponding SHA hash to check file integrity
        @param\tcheck_SHA\tPython bool[default=True]: whether to check the file integrity
        @params\tquiet\tPython bool[default=False]
        @return\t0\tmeaning that the download was successful
    '''
    if (not os.path.exists(path_to_lincs+file_name)):
        if (not os.path.exists(path_to_lincs+file_name+".gz")):
            cmd = "wget -c "+("" if (not quiet) else "-q ")+"-O "+path_to_lincs+file_name+".gz "+base_url+file_name+".gz"
            sbcall(cmd, shell=True)
            if (check_SHA):
                ## Checks file integrity
                sha_table = pd.read_csv(path_to_lincs+file_sha, sep="  ", names=["sha", "file"], engine='python')
                true_sha_id = list(sha_table[sha_table["file"] == file_name+".gz"]["sha"])[0]
                sha_id = sb.check_output("sha512sum "+path_to_lincs+file_name+".gz", shell=True).decode("utf-8").split("  ")[0]
                assert sha_id == true_sha_id
        sbcall("gzip -df "+path_to_lincs+file_name+".gz", shell=True)
        if (not quiet):
            print("<LINCS_utils> "+file_name+" successfully downloaded in "+path_to_lincs)
    else:
        ## Checks file integrity
        if (check_SHA and os.path.exists(path_to_lincs+file_name+".gz")):
            sha_table = pd.read_csv(path_to_lincs+file_sha, sep="  ", names=["sha", "file"], engine='python')
            true_sha_id = list(sha_table[sha_table["file"] == file_name+".gz"]["sha"])[0]
            sha_id = sb.check_output("sha512sum "+path_to_lincs+file_name+".gz", shell=True).decode("utf-8").split("  ")[0]
            if (not (sha_id == true_sha_id)):
                cmd = "wget "+("" if (not quiet) else "-q ")+"-c -O "+path_to_lincs+file_name+".gz "+base_url+file_name+".gz"
                sbcall(cmd, shell=True)
                if (check_SHA):
                    ## Checks file integrity
                    sha_table = pd.read_csv(path_to_lincs+file_sha, sep="  ", names=["sha", "file"], engine='python')
                    true_sha_id = list(sha_table[sha_table["file"] == file_name+".gz"]["sha"])[0]
                    sha_id = sb.check_output("sha512sum "+path_to_lincs+file_name+".gz", shell=True).decode("utf-8").split("  ")[0]
                    assert sha_id == true_sha_id
                sbcall("gzip -df "+path_to_lincs+file_name+".gz", shell=True)
                if (not quiet):
                    print("<LINCS_UTILS> "+file_name+" successfully (re)downloaded in "+path_to_lincs)
    return 0

def download_lincs_files(path_to_lincs, which_lvl):
    '''
        Returns and downloads the proper LINCS L1000 files from Gene Expression Omnibus (GEO)
        @param\tpath_to_lincs\tPython character string: path to folder in which LINCS L1000-related files will be locally stored
        @param\twhich_lvl\tPython integer list: LINCS L1000 Level to download (either [3] -normalized gene expression-, [5] -binary experimental signatures-, [3,5])
        @return\tfile_list\tList of 4 Python character string lists: gene_files, sig_files, lvl3_files, lvl5_files Python lists of character strings
    '''
    assert all([x in [3,5] for x in which_lvl])
    ## Latest versions of LINCS L1000 as of September 2022
    lincs_gse = {
		"phase1": {
			"acc": "GSE92742",
			"file_lvl5": "Level5_COMPZ.MODZ_n473647x12328.gctx",
			"file_lvl3": "Level3_INF_mlr12k_n1319138x12328.gctx",
			"gene_file": "gene_info.txt",
			"sig_file": "sig_info.txt"
		},
		"phase2": {
			"acc": "GSE70138",
			"file_lvl5": "Level5_COMPZ_n118050x12328_2017-03-06.gctx",
			"file_lvl3": "Level3_INF_mlr12k_n345976x12328_2017-03-06.gctx",
			"gene_file": "gene_info_2017-03-06.txt",
			"sig_file": "sig_info_2017-03-06.txt"
		}
	}
    keys = ["gene_file", "sig_file", "file_lvl3", "file_lvl5"]
    file_di = {}
    for k in keys:
        file_di.setdefault(k, [])
    for key in list(lincs_gse.keys()):
        acc = lincs_gse[key]["acc"]
        base_url = "ftp://ftp.ncbi.nlm.nih.gov/geo/series/"+acc[:5]+"nnn/"+acc+"/suppl/"
        for k in keys:
            file_di[k] = file_di[k]+[acc+"_Broad_LINCS_"+lincs_gse[key][k]]
        file_sha = acc+"_SHA512SUMS.txt"
        download_file(path_to_lincs, file_sha, base_url, None, check_SHA=False)
        # 5GB! ~20 minutes with a good Internet connection (level 5)
        # 48.8GB and 12.6GB! (level 3)
        for lvl in which_lvl:
            download_file(path_to_lincs, file_di["file_lvl"+str(lvl)][-1], base_url, file_sha)
        for k in keys[:2]:
            download_file(path_to_lincs, file_di[k][-1], base_url, file_sha)	
    return [file_di[k] for k in keys]

####################
## RETRIEVE DATA  ##
####################

def create_restricted_drug_signatures(sig_ids, entrezid_list, path_to_lincs, which_lvl=[3], strict=True, quiet=False):
    '''
        Create dataframe of drug signatures from LINCS L1000 from a subset of signature and gene IDs
        @param\tsig_ids\tPython character string list: list of signature IDs from LINCS L1000 (Level 3: "distil_id", Level 5: "sig_id")
        @param\tentrezid_list\tPython character string list: list of EntrezIDs 
        @param\tpath_to_lincs\tPython character string: folder in which LINCS L1000-related files are stored
        @param\twhich_lvl\tPython integer list: [3] for Level 3, [5] for Level 5 
        @param\tstrict\tPython bool[default=True]: if set to True, if not all signatures are retrieved, then return None. If set to False, return the (sub)set of retrievable signatures
        @param\tquiet\tPython bool[default=False]
        @return\tsigs\tPandas DataFrame: rows/[] x columns/[]
    '''
    assert all([x in [3,5] for x in which_lvl])
    assert len(which_lvl)==1
    gene_files, sig_files, lvl3_files, lvl5_files = download_lincs_files(path_to_lincs, which_lvl=which_lvl)
    sigs = []
    cid = ("sig_id" if (5 in which_lvl) else "distil_id")
    for idf, sf in enumerate(sig_files):
        df = pd.read_csv(path_to_lincs+sf, sep="\t", engine='python')
        if (3 in which_lvl):
            sig_selection = [sig for sig in list(set(sig_ids)) if any([sig in x.split("|") for x in list(df[cid])])]
        else:
            sig_selection = [sig for sig in list(set(sig_ids)) if (sig in list(df[cid]))]
        if (len(sig_selection) > 0):
            df = pd.read_csv(path_to_lincs+gene_files[idf], sep="\t", engine='python', dtype=str)
            if (entrezid_list is None):
                gene_selection = list(df.index)
            else:
                gene_selection = [g for g in list(df["pr_gene_id"]) if (int(g) in entrezid_list)]
            try:
                # https://github.com/cmap/cmapPy/blob/master/cmapPy/pandasGEXpress/parse_gctx.py
                dataset = parse_gctx.parse(path_to_lincs+(lvl5_files if (5 in which_lvl) else lvl3_files)[idf], cid=sig_selection, rid=gene_selection).data_df
            except Exception as e:
                # means "not in metadata for the considered GCTX file"
                print("<LINCS> "+str(e))
                continue
            dataset = dataset.loc[gene_selection][sig_selection]
            dataset.index = [int(x) for x in gene_selection]
            assert dataset.shape[0]==len(gene_selection)
            assert dataset.shape[1]==len(sig_selection)
            sigs.append(dataset)
    if (len(sigs)>0):
        sigs = sigs[0].join(sigs[1:], how="outer")
    else:
        if (not quiet):
            print("\n<LINCS> /!\ None of the following signature/profile ids was found -> {"+"  ".join(sig_ids)+"}")
        if (strict):
            raise ValueError("<LINCS> No signature/profile found.")
        else:
            return None
    assert (len(sig_ids)==sigs.shape[1]) or (not strict)
    if (len(sigs)>0 and ((len(sigs.columns) < len(sig_ids)) and strict)):
        if (not quiet):
            print("\n<LINCS> /!\ Some of the following signature/profile ids were not found -> {"+"  ".join([x for x in sig_ids if (x not in sigs.columns, sig_ids)])+"}")
        return None
    return sigs

def select_best_sig(params, filters, user_key, selection="distil_ss", nsigs=2, same_plate=True, iunit=None, quiet=False):
    '''
        Select "best" set of profiles ("experiment") (in terms of quality, or criterion "selection") according to filters
        @param\tparams\tPython dictionary: additional arguments for the request
        @param\tfilters\tPython dictionary: additional arguments for filtering the results of the request (defined with params)
        @param\tselection\tPython character string[default="distil_ss"]: name of the metric in LINCS L1000 to define the best signature
        @param\tnsigs\tPython integer[default=2]: minimum number of signatures to retrieve
        @param\tsame_plate\tPython bool[default=True]: retrieve signatures from the same plate or not
        @param\tiunit\tPython character string
        @param\tquiet\tPython bool[default=False]
        @return\tdata\tPython dictionary list:
    '''
    assert len(selection) == 0 or (selection in list(filters.keys()) or selection in params.get("fields", []))
    id_fields = ["brew_prefix"]
    if (any([not f in params.get("fields", []) for f in id_fields]) or (len(params.get("fields", [])) == 0 or (len(filters) > 0 and any([not (f in params.get("fields", [])) for f in list(filters.keys())])))):
        fields = params.get("fields", [])
        fields += list(filters.keys())
        params["fields"] = list(set(fields+id_fields))
    endpoint = "sigs"
    method = "filter"
    request_url = build_url(endpoint, method, params=params, user_key=user_key)
    ## GET EXPERIMENTS MATCHING THE PARAMS
    data = post_request(request_url, quiet=True, pause_time=0.3)
    ## 1. Test if enough signatures were found
    if (nsigs > 0 and (len(data) == 0 or (same_plate and all([len(d["distil_id"]) < nsigs for d in data])) or (not same_plate and sum([len(d["distil_id"]) for d in data]) < nsigs))):
        raise ValueError("<LINCS> (1) No (enough) signatures (%d instead of min. %d) retrieved via LINCS L1000 API.\n%s\n" % (sum([len(d["distil_id"]) for d in data]), nsigs, request_url))
    if (not quiet):
        if ("where" not in params):
            cond = ""
        else:
            if ("pert_type" in params["where"]):
                cond = "TREATED" if ("trt_" == params["where"]["pert_type"][:4]) else "CONTROL"
            else:
                cond = ""
        print("<LINCS> %d %s experiments ('brew_prefix') (%d profiles 'distil_id' in average)" % (len(data), cond, np.mean([len(d["distil_id"]) for d in data])))
    ## 2. Filter signatures according to the filters
    if (len(filters) > 0):
        data_ = [x for x in data if (all([x.get(k, filters[k]+1) > filters[k] for k in list(filters.keys())]))]
        if (not quiet):
            print("<LINCS> %d/%d filtered experiments" % (len(data_),len(data)))
        data = data_
    ## 3. Check if matches the specified (if exists) dose
    if ("pert_dose_unit" in params["fields"] and iunit):
        data = [x for x in data if (x["pert_dose_unit"] == iunit.decode("utf-8"))]
    ## 4. Keep only valid profile IDs ('distil_id')
    if ("distil_id" in params["fields"]):
        data_ = []
        for i in range(len(data)):
            distil_ids = [x for x in data[i]["distil_id"] if (len(x.split(":")) == 2)]
            if (len(distil_ids) > 0):
                d = data[i]
                d["distil_id"] = distil_ids
                data_.append(d)
        data = data_
    ## 5. Select the best set of profiles ("experiment") based on the criterion "selection"
    if (same_plate and "distil_id" in params["fields"] and len(data) > 0):
        data = [x for x in data if (len(list(set(x["distil_id"]))) >= nsigs)]
        argmax_rank = int(np.argmax([float(x.get(selection, 0)) for x in data]))
        data_rank = data[argmax_rank]
        if (not quiet):
            print("<LINCS> Selected "+selection+" = "+str(data_rank[selection]))
        data = [{"distil_id": distil_id} for distil_id in list(set(data_rank["distil_id"]))]
        for f in [x for x in params["fields"] if (x != "distil_id")]:
            for i in range(len(data)):
                data[i].setdefault(f, data_rank[f])
        if (not quiet):
            print("<LINCS> %d same-plate profiles" % len(data))
    elif ("distil_id" in params["fields"]):
        for i in range(len(data)):
            data[i]["distil_id"] = data[i]["distil_id"][0]
    return data

def compute_interference_scale(sigs, samples, entrez_id, is_oe, taxon_id, lincs_specific_ctl_genes, quiet=True, eps=2e-7):
    '''
        Computes the interference scale [1] which determines whether a genetic perturbation was successful
	@param\tsigs\tPandas DataFrame: rows/[genes] x columns/[control and treated samples]
	@param\tsamples\tPython integer list: contains 1 for control samples, 2 for treated ones for each column of @sigs
	@param\tentrez_id\tPython integer: EntrezID of the perturbed gene 
	@param\tis_oe\tPython bool: is the experiment an overexpression of the perturbed gene (is_oe=True) or a knockdown
	@param\tquiet\tPython bool[default=True]
	@param\teps\tPython float[default=2e-7]
	@return\tiscale\tPython float: interference scale for the input experiment
        [1] doi.org/10.1002/psp4.12107 
    '''
    assert len(sigs.columns) == len(samples)
    assert all([s in [1,2] for s in samples])
    assert entrez_id in sigs.index
    ## "Housekeeping genes" selected in PMC5192966
    treated = sigs.columns[np.array(samples)==2]
    control = sigs.columns[np.array(samples)==1] 
    exp_ratio = np.mean(sigs[treated].loc[entrez_id])/float(np.mean(sigs[control].loc[entrez_id])+eps)
    ctl_genes = [g for g in list(lincs_specific_ctl_genes.values()) if (g in sigs.index)]
    assert len(ctl_genes) > 0
    hk_inv_ratios = [np.mean(sigs[control].loc[g])/float(np.mean(sigs[treated].loc[g])+eps) for g in ctl_genes]
    ## The most stable (ratio ~ 1) between control and treated groups
    best_hk_gene_id = int(np.argmin(np.abs(np.array(hk_inv_ratios)-1)))
    iscale = exp_ratio*hk_inv_ratios[best_hk_gene_id]
    assert not pd.isna(iscale)
    iscale = iscale-1 if (is_oe) else 1-iscale
    if (not quiet):
        print("<INTERFERENCE SCALE> %s Perturbed %d = %.5f || Most stable control %s = %.5f" % ("OE" if (is_oe) else "KD", entrez_id,exp_ratio,ctl_genes[best_hk_gene_id],1./hk_inv_ratios[best_hk_gene_id]))
        print("<INTERFERENCE SCALE> %.5f" % iscale)
    return iscale

def get_treated_control_dataset(treatment, pert_type, cell, filters, entrez_ids, taxon_id, user_key, path_to_lincs, entrez_id=None, selection="distil_ss", dose=None, iunit=None, itime=None, which_lvl=[[3,5][1]], nsigs=2, same_plate=True, quiet=False, trim_w_interference_scale=True, return_metrics=[]):
    '''
        Retrieve set of experimental profiles, with at least @nsigs treated and control sample
        @param\ttreatment\tPython character string: HUGO gene symbol
        @param\tpert_type\tPython character string: type of perturbation as accepted by LINCS L1000
        @param\tcell\tPython character string: cell line existing in LINCS L1000
        @param\tfilters\tPython dictionary: additional parameters for the LINCS L1000 requests
        @param\tentrez_ids\tPython integer list: EntrezID genes
        @param\ttaxon_id\tPython integer: NCBI taxonomy ID
        @param\tuser_key\tPython character string: LINCS L1000 user API key
        @param\tpath_to_lincs\tPython character string: path where LINCS L1000 files are locally stored
        @param\tentrez_id\tPython integer: EntrezID identifier for HUGO gene symbol @treatment
        @param\tselection\tPython character string[default="distil_ss"]: LINCS L1000 metric which is maximized by a given experiment
        @param\tdose\tPython character string or None[default=None]
        @param\tiunit\tPython character string or None[default=None]
        @param\titime\tPython character string or None[default=None]
        @param\twhich_lvl\tPython integer list[default=[3]]
        @param\tnsigs\tPython integer[default=2]: minimal number of samples of each condition in each experiment
        @param\tsame_plate\tPython bool[default=True]: select samples from the same plate for each experiment and condition
        @param\tquiet\tPython bool[default=True]
        @param\ttrim_w_interference_scale\tPython bool[default=True]: computes the interference scale criteria for further trimming
        @param\treturn_metrics\tPython character string list: list of LINCS L1000 metrics to return as the same time as the profiles
        @return\tsigs\tPandas DataFrame: rows/[genes+"annotation"+"signame"+"sigid"] x columns/[profiles] or None
    '''
    assert all([x in [3,5] for x in which_lvl])
    assert len(which_lvl)==1
    assert user_key
    assert nsigs > 1
    if (pert_type == "trt_cp"):
        trim_w_interference_scale = False
    lincs_specific_ctl_genes = convert_ctrlgenes_EntrezGene(taxon_id)
    genes, genes_controls = list(set(entrez_ids)), list(set(list(lincs_specific_ctl_genes.values())+entrez_ids))
    ## 1. Build LINCS L1000 request to determine IDs of suitable experimental profiles
    where = {"cell_id": cell, "pert_type": pert_type}
    if (len(treatment) > 0):
        where.setdefault("pert_iname", treatment)
    if (dose is not None):
        where.setdefault("pert_dose", dose)
    if (itime is not None):
        where.setdefault("pert_itime", itime)
    # https://www.biostars.org/p/211896/
    cid = ("sig_id" if (which_lvl == 5) else "distil_id")
    fields = list(set(list(filters.keys())+[cid]+([selection] if (len(selection) > 0) else [])+return_metrics))
    params = { "where": where, "fields": fields+["pert_dose_unit"] }
    ## 2. Retrieve IDs of treated experiments (maximizing criterion @selection)
    data_treated = select_best_sig(params, filters, user_key, selection=selection, nsigs=nsigs, same_plate=same_plate, quiet=quiet, iunit=iunit)
    assert len(data_treated) > 0
    where = {"cell_id": cell, "pert_type": ("ctl_vehicle" if (pert_type == "trt_cp") else "ctl_vector")}
    if (same_plate):
        where.setdefault("brew_prefix", str(data_treated[0]["brew_prefix"][0]))
    ## 3. Retrieve IDs of associated control experiments 
    params = { "where": where, "fields": fields }
    data_control = select_best_sig(params, filters, user_key, selection=selection, nsigs=nsigs, same_plate=same_plate, quiet=quiet)
    assert len(data_control) > 0
    ## 4. Collect all IDs of treated and control experimental profiles 
    sig_ids = [str(x[cid]) for x in data_treated+data_control]
    sigs = create_restricted_drug_signatures(sig_ids, genes if (not trim_w_interference_scale) else genes_controls, path_to_lincs, which_lvl=which_lvl, strict=True)
    assert len(sigs) > 0
    ## 5. Get annotations (control/treated, profile/signature IDs) for each experimental profile
    signame_trt, signame_ctl = [[str(s[cid]) for s in x] for x in [data_treated, data_control]]
    samples = [2*int(c in signame_trt)+int(c in signame_ctl) for c in sigs.columns]
    assert all([s in [1,2] for s in samples])
    sigs.loc["annotation"] = samples
    ## 6. If returning other metrics, append them to the Pandas DataFrame
    if (len(return_metrics) > 0):
        dtc = data_treated+data_control
        metric_vals = pd.DataFrame([[x.get(metric, np.nan) for metric in return_metrics] for x in dtc], columns=sig_ids, index=return_metrics)
        sigs = pd.concat((sigs, metric_vals), axis=0)
    ## 7. Compute interference scale for genetic experiments if needed for further trimming
    if (trim_w_interference_scale):
        if (entrez_id not in sigs.index):
            return None
        select = [col for col in sigs.columns if (sigs.loc["annotation"][col] in [1, 2])]
        iscale = compute_interference_scale(sigs[select].drop(["annotation"]), sigs[select].loc["annotation"], entrez_id, (pert_type in ["trt_oe"]), taxon_id, lincs_specific_ctl_genes, quiet=quiet)
        sigs.loc["interference_scale"] = [iscale]*sigs.shape[1] ## common to all profiles from the same experiment (treated/control)
    sigs = sigs.loc[[g for g in genes if (g in sigs.index)]+["annotation"]+return_metrics+([] if (not trim_w_interference_scale) else ["interference_scale"])]
    return sigs



# coding: utf-8

complexity_limit=1e6

import gc
import numpy as np
from time import time

import NORDic.NORDic_DR.utils as utils

bandit_types = ["MisLid", "LinGapE"]
beta_types = ["heuristic", "linear", "lucb1", "misspecified", "gaussian", "subheuristic"]
learner_types = ["AdaHedge", "Greedy"]

####################################################
## Thresholds                                     ##
####################################################

# "Iterated log" inspired threshold [Kaufmann et al., 2015]
def heuristic(X, delta, sigma, c):
    N, K = X.shape
    def f(t, na, x=None):
        return np.log((1+np.log(t+1))/float(delta))
    return f

def subheuristic(X, delta, sigma, c):
    N, K = X.shape
    def f(t, na, x=None):
        return np.log(t+1)/float(delta)
    return f

# This paper's stopping rule for misspecified linear models
def misspecified(X, delta, sigma, c):
    N, K = X.shape
    L = float(np.max(np.linalg.norm(X, axis=0)))
    e = np.exp(1)
    def f(t, na, x=None):
        if (str(x) == "None"):
            x = 2*L
        term_lin = (lambda t : 1+np.log(2/float(delta))+(1+1/np.log(2/float(delta)))*0.5*N*np.log(1+t*L**2/(x*N)*np.log(2/float(delta)))+2*c**2*t)
        term_uns = (lambda t : 2*K*utils.lambert(1/float(2*K)*np.log(2*e/float(delta))+0.5*np.log(8*e*K*np.log(t))))
        return min(term_lin(t), term_uns(t))
    return f

# [Kaufmann et Koolen, 2018]
def gaussian(X, delta, sigma, c):
    ## S subset of arms
    S = range(X.shape[1])
    K = len(S)
    from scipy.special import zeta
    gC = lambda l : 2*l-2*l*np.log(4*l)+np.log(zeta(2*l, q=1))-0.5*np.log(1-l) 
    from scipy.optimize import minimize
    x = np.log(1./float(delta))/float(K)
    fc = lambda l : (gC(l)+x)/float(l)
    lam = minimize(fc, np.array([0.75]), bounds=[(0.511111,0.9999)], tol=1e-6, options={"maxiter":1000}).x
    Cx = fc(lam)
    def f(t, na, x=None):
        return float(2*np.sum(np.log(4+np.log(np.array(na)+np.asarray(np.array(na)==0, dtype=int))))+K*Cx)
    return f

####################################################
## Learners                                       ##
####################################################

# Learner structure
class Learner(object):
	def __init__(self, K, name="UnimplementedLearner"):
		self.K, self.name = K, name
		## Arm playing mixed strategy
		self.p = np.zeros(self.K)
	def act(self):
		raise NotImplemented
	def incur(self, w):
		raise NotImplemented

# Best Response learner
class Greedy(Learner):
	def __init__(self, K):
		super(Greedy, self).__init__(K, name="FTL")
	def act(self):
		p = np.asarray(self.p == np.min(self.p), dtype=float)
		p /= np.sum(p)
		return p
	def incur(self, w):
		self.p[:] = w

# AdaHedge learner (the one used in LinGame)
class AdaHedge(Learner):
	def __init__(self, K):
		self.delta = 0.01
		super(AdaHedge, self).__init__(K, name="AdaHedge")

	def act(self):
		eta = np.log(self.K)/float(self.delta)
		p = np.exp(-eta*(self.p-np.min(self.p)))
		p /= np.sum(p)
		return p

	def incur(self, w):
		p = self.act()
		eta = np.log(self.K)/float(self.delta)
		self.p += w
		m = np.min(w)-1./eta*np.log(p.T.dot(np.exp(-eta*(w-np.min(w)))))
		assert m != float("inf") and p.T.dot(w) >= m-1e-7
		self.delta += p.T.dot(w) - m

####################################################
## Bandit algorithm                               ##
####################################################

## Generic class
class Misspecified(object):
    def __init__(self, method_args):
        for k in method_args:
            setattr(self, k, method_args[k])

    def clear(self):
        self.rewards = []
        self.samples = []
        self.t = 0
        self.na = []

    def sample(self, problem, candidates):
        if (len(self.na)==0):
            self.na = [0]*(problem.X.shape[1])
        for arm in candidates:
            self.rewards.append(problem.reward(arm))
            self.samples.append(arm)
            self.t += 1
            self.na[arm] += 1

    def run(self, problem, nsimu, run_id=None, quiet=False):
        active_arms = np.array([0]*(problem.X.shape[1]))
        complexity = [None]*nsimu
        running_time = [None]*nsimu
        for simu in range(nsimu):
            self.clear()
            starting_time = time()
            active_arms_, complexity_ = self.apply(problem)
            running_time[simu] = time()-starting_time
            complexity[simu] = complexity_
            active_arms[active_arms_] += 1
            C = round(np.mean(complexity[:simu+1]),2)
            id_ = simu + 1 if run_id is None else run_id
            if (not quiet):
                print("It. #%d: %d samples (running avg. %d)" % (id_, complexity_, C))
            gc.collect()
        return active_arms/float(nsimu), complexity, running_time

## For misspecified models
class MisLid(Misspecified):
    def __init__(self, method_args):
        self.name = "MisLid"
        if (method_args["learner_name"] == "AdaHedge"):
            self.learner_type = AdaHedge
        elif (method_args["learner_name"] == "Greedy"):
            self.learner_type = Greedy
        else:
            raise ValueError("Learner '"+method_args["learner_name"]+"' not implemented.")
        self.T_init = 1 
        assert method_args["sigma"] == 1
        assert "beta_linear" in method_args
        method_args["beta"] = method_args["beta_linear"]
        self.m = method_args["m"]
        assert method_args["sigma"] == 1.
        self.lambda_ = method_args["sigma"]/float(20.)
        assert self.lambda_ > 0
        self.constraint = "L_inf"
        assert self.constraint in ["L_inf", "L1"]
        self.multi_learners, self.M = 0, 1
        self.cnorm = lambda x : utils.cnorm(x, norm=self.constraint)
        super(MisLid, self).__init__(method_args)

    def stopping_rule(self, quiet=True):
        if (not quiet):
            print("t="+str(self.t)+" B(t)="+str(self.B)+" thres(t)="+str(self.epsilon)+" stop="+str(self.B > self.epsilon))
        return (self.B > self.epsilon) or (self.t > complexity_limit)

    def best_answer(self, means):
        J = utils.argmax_m(means, self.m)
        return J

#    def update(self, problem, candidates, Vinv, b, Vinv_val=None):
#        return utils.update_misspecified(problem, candidates, Vinv, b, self.c, self.na, self.rewards, Vinv_val=Vinv_val)

    def update(self, problem, candidates, Vinv, b, Vinv_val=None):
        candidates.reverse()
        for ia, a in enumerate(candidates):
            i_a = -(ia+1)
            Vinv = utils.sherman_morrison(Vinv, problem.X[:, a].reshape((problem.X.shape[0], 1)))
            b += self.rewards[i_a]*problem.X[:, a].reshape(b.shape)
        theta = Vinv.dot(b)
        means = np.array(theta.T.dot(problem.X)).flatten().tolist()
        eta = np.zeros((problem.X.shape[1], 1))
        return Vinv, b, means, theta, eta

    def apply(self, problem, precision=1e-7):
        self.B, self.epsilon = -float("inf"), 0.
        N, K = problem.X.shape
        means = [0]*K
        b = np.zeros((N,1))
        learners, sum_w = {}, np.zeros(K)
        stop = False
        L = float(np.max([self.cnorm(problem.X[:,a]) for a in range(K)]))
        self.x = 2*L

        A = problem.X.T
        print("minimum eigenvalue of A^TA: {}".format(np.linalg.eigvals(A.T.dot(A)).min()))
        
        ## Consider a (C-approximate) barycentric spanner
        F = utils.barycentric_spanner(problem.X, C=1, quiet=True)
        ## Uncomment the next two lines to get regularized version of design matrix instead
        #F = range(K)
        #self.T_init = 0
        candidates, maxiter = [], len(F)*(self.T_init+1)
        from scipy.sparse.linalg import eigsh
        get_eigmin = lambda M : eigsh(M, k=1, which="SM")[0][0]
        V = np.zeros((N,N))
        alternative_arms = []  # arms in the recent closest alternatives
        stopping_rule_time = max(5*N, 10)  # next time at which the sampling rule is checked
        while (self.t < maxiter): # t0 initialization phases to check V_t0 >= x Id
            ## Round-Robin procedure
            for a in F:
                self.sample(problem, [a])
                V += problem.X[:,a].dot(problem.X[:,a].T)
            candidates += F
            sign, logdet = np.linalg.slogdet(V-self.x*np.eye(N))
            detVxId = sign*np.exp(logdet)
            sign, logdet = np.linalg.slogdet(V)
            detV = sign*np.exp(logdet)
            if (detVxId>=0 and detV>0):
                break    
        if (self.T_init > 0):
            Vinv = np.linalg.pinv(V)
            Vinv, b, means, theta, eta = self.update(problem, candidates, Vinv, b, Vinv_val=Vinv)
        else:
            ## Regularized version (necessary for invertible design matrices)
            ## In practice, never used
            Vinv = 1/float(self.lambda_**2)*np.eye(N)
            Vinv, b, means, theta, eta = self.update(problem, candidates, Vinv, b)
            V = np.linalg.pinv(Vinv)
            self.x = get_eigmin(V)
        while (not stop):
            S_t = self.best_answer(means)  # current best answer set
            
            if (self.multi_learners):  # Several learners
                learner = learners.get(tuple(S_t), None)  # get step from learner associated with best answer
            else:  # Only one learner
                learner = learners.get(0, None)  # get step from learner associated with best answer
            if (str(learner) == "None"):
                learner = self.learner_type(K)
            
            if (self.learner_name == "AdaHedge"):
                # query the learner
                w_t = learner.act() 
                # get closest alternative to current means - Best-Response
                means_alt, closest, val, alternative_arms = utils.closest_alternative(
                    problem, b, means, theta, eta, w_t, self.c,  S_t,
                    constraint=self.constraint, subsample=self.subsample, alternative_arms=alternative_arms)
            elif (self.learner_name == "Greedy"):
                # get closest alternative to current means - FTL
                means_alt, closest, val, alternative_arms = utils.closest_alternative(
                    problem, b, means, theta, eta, self.na, self.c,  S_t,
                    constraint=self.constraint, subsample=self.subsample, alternative_arms=alternative_arms)
            else:
                raise ValueError("Unknown learner type.")

            # optimism
            mu = np.array(means).flatten()
            lambda_ = np.array(means_alt).flatten()
            delta = -utils.optimistic_gradient(problem, Vinv, mu, lambda_, self.na, self.t, self.M, self.c,
                self.gain_type if (str(self.gain_type)!="None") else "misspecified")
            # update learner
            learner.incur(delta)  
            di_learner = {}
            di_learner.setdefault(tuple(S_t) if (self.multi_learners) else 0, learner)

            if (self.learner_name == "Greedy"):
                # pull argmin of the loss
                w_t = learner.act()

            nb_samples = 1
            if self.subsample:
                nb_samples = N
            for sample_id in range(nb_samples):  # Sample the same arm several times
                sum_w += w_t
                learners.update(di_learner)
                candidate = utils.tracking_rule(w_t, sum_w, self.na, self.t, self.tracking_type if (str(self.tracking_type)!="None") else "S") # default is "S"
                self.sample(problem, candidate)
                # update sum of rewards, estimated parameters and means
                Vinv, b, means, theta, eta = self.update(problem, [candidate], Vinv, b)
            # Stopping rule test
            if (((self.t > stopping_rule_time) and self.subsample) or (not self.subsample)):
                means_alt, closest, val, alternative_arms = utils.closest_alternative(problem, b, means, theta, eta, self.na,
                    self.c, self.best_answer(means), constraint=self.constraint, subsample=False, alternative_arms=alternative_arms)
                self.B = val
                self.epsilon = self.beta(self.t, self.na, x=self.x)
                stop = self.stopping_rule(quiet=(self.t % 1000 != 0))  # check stopping rule
                #stop = self.stopping_rule(quiet=False)
                stopping_rule_time = max(self.t, stopping_rule_time)*self.geometric_factor
        import pandas as pd
        print(pd.DataFrame([
	        means,
	        [np.mean(np.array(self.rewards)[np.array(self.samples)==a]) for a in np.unique(self.samples)]
        ],index=["means", "emp. means"], columns=problem.targets.columns))
        J = self.best_answer(means)
        return J, self.t

class LinGapE(Misspecified):
    def __init__(self, method_args):
        self.T_init = 1
        assert "beta_linear" in method_args
        self.name = "LinGapE"
        method_args["beta"] = method_args["beta_linear"]
        super(LinGapE, self).__init__(method_args)

    def stopping_rule(self):
        if (self.t % 2000 == 0):
            print("t="+str(self.t)+" B(t) = "+str(self.B))
        return (self.B <= self.epsilon) or (self.t > complexity_limit)

    def greedy(self, problem, b_t, c_t, Vinv):
        K = problem.X.shape[1]
        direction = problem.X[:, b_t]-problem.X[:, c_t]
        uncertainty = [float(utils.mahalanobis(direction.reshape((problem.X.shape[0],1)), utils.sherman_morrison(Vinv, problem.X[:,i].reshape((problem.X.shape[0],1))))) for i in range(K)]
        a = utils.randf(uncertainty, 1, np.min)[0]
        return [a]

    def optimized(self, problem, b_t, c_t, Vinv):
        p = self.ratio.get((b_t,c_t))
        if (not len(p)):
            ## https://math.stackexchange.com/questions/1639716/how-can-l-1-norm-minimization-with-linear-equality-constraints-basis-pu
            from scipy.optimize import linprog
            X = problem.X
            Aeq = np.concatenate((X, -X), axis=1)
            beq = (X[:,b_t]-X[:,c_t])
            F = np.ones((2*K, ))
            bounds = [(0, float("inf"))]*(2*K)
            solve = linprog(F, A_eq=Aeq, b_eq=beq, bounds=bounds)
            x = solve.x
            w = x[:K]-x[K:]
            assert solve.status == 0
            p = np.abs(w)
            p /= np.linalg.norm(w, 1)
            self.ratio.setdefault((b_t,c_t), p)
        samplable_arms = [i for i in range(problem.X.shape[1]) if (float(p[i]) > 0)]
        a = samplable_arms[self.randf([float(self.na[i]/float(p[i])) for i in samplable_arms], 1, np.min)]
        return [a]

    def update(self, problem, candidates, Vinv, b):
        candidates.reverse()
        for ia, a in enumerate(candidates):
            i_a = -(ia+1)
            Vinv = utils.sherman_morrison(Vinv, problem.X[:, a].reshape((problem.X.shape[0], 1)))
            b += self.rewards[i_a]*problem.X[:, a].reshape(b.shape)
        theta = Vinv.dot(b)
        means = np.array(theta.T.dot(problem.X)).flatten().tolist()
        return Vinv, b, means

    def apply(self, problem, greedy_sampling=True, lambda_val=1.):
        self.B = float("inf")
        self.ratio = {}
        N, K = problem.X.shape
        Vinv = 1/float(lambda_val**2)*np.matrix(np.eye(N)).reshape((N,N))
        b = np.matrix(np.zeros(N)).reshape((N,1)) 
        means = [0]*K
        self.Cbeta = lambda t, n : np.sqrt(2*self.beta(t,n))
        w = lambda t, c, a, Vinv : float(self.Cbeta(t, self.na[c]+self.na[a])*utils.mahalanobis((problem.X[:,c]-problem.X[:,a]).reshape((problem.X.shape[0],1)), Vinv))
        w_ind = lambda t, a, Vinv : float(self.Cbeta(t, self.na[a])*utils.mahalanobis(problem.X[:,a].reshape(problem.X.shape[0],1), Vinv))
        Bidx = lambda t, i, j, Vinv : float(means[i]-means[j]+w(t, i, j, Vinv))
        stop = False
        for _ in range(self.T_init):
            for a in range(K):
                self.sample(problem, [a])
        if (self.T_init > 0):
            Vinv, b, means = self.update(problem, [a for _ in range(self.T_init) for a in range(K)], Vinv, b)
        while (not stop):
            J = utils.argmax_m(means, self.m)
            notJ = [a for a in range(K) if (a not in J)]
            indices = [[means[c]-means[a]+w(self.t, c, a, Vinv) for c in notJ] for a in J]
            max_idx = utils.randf([np.max(ids) for ids in indices], 1, np.max)[0]
            b_t = J[max_idx]
            max_idx_idx = utils.randf(indices[max_idx], 1, np.max)[0]
            c_t = notJ[max_idx_idx]
            if (greedy_sampling):
                candidate = self.greedy(problem, b_t, c_t, Vinv)
            else:
                candidate = self.optimized(problem, b_t, c_t, Vinv)
            self.sample(problem, candidate)
            Vinv, b, means = self.update(problem, candidate, Vinv, b)
            self.B = float(indices[max_idx][max_idx_idx])
            stop = self.stopping_rule()
        J = utils.argmax_m(means, self.m)
        return J, self.t


#coding:utf-8

from subprocess import call as sbcall
import pandas as pd
import numpy as np
import omnipath as op

from NORDic.UTILS.STRING_utils import get_interactions_from_STRING, get_network_from_STRING
from NORDic.UTILS.utils_data import request_biodbnet
from NORDic.UTILS.utils_grn import get_weakly_connected

import contextlib
@contextlib.contextmanager
def capture():
    import sys
    from io import StringIO
    oldout,olderr = sys.stdout, sys.stderr
    try:
        out=[StringIO(), StringIO()]
        sys.stdout,sys.stderr = out
        yield out
    finally:
        sys.stdout,sys.stderr = oldout, olderr
        out[0] = out[0].getvalue()
        out[1] = out[1].getvalue()

def determine_edge_threshold(network, core_gene_set, quiet=True):
    '''
        Determine the greatest threshold on the edge score which allows all of the core gene set to be connected (binary search)
        @param\tnetwork\tPandas DataFrame: rows/[interactions] x at least three columns "preferredName_A" (input node), "preferredName_B" (output node), "score" (edge score)
        @param\tcore_gene_set\tPython character string list: list of genes that should remain connected
        @param\tquiet\tPython bool[default=None]: 
        @return\tt\tPython float: maximum threshold which allows the connection of all genes in the core set
    '''
    ppi = network[["preferredName_A","preferredName_B","score"]]
    ppi.columns = ["A","B","score"]
    glist = list(set(list(ppi["A"])+list(ppi["B"])))
    components = get_weakly_connected(ppi, glist, index_col="A", column_col="B", score_col="score")
    main_component = [g for g in components[0] if (g in core_gene_set)]
    if (not quiet):
        print("Connected core gene set of size %d" % len(main_component))
    isolated_genes = [g for c in components[1:] for g in c]
    if (len(isolated_genes)>0):
        test_notisolated = np.vectorize(lambda x : x not in isolated_genes)
        ppi = ppi.loc[test_notisolated(ppi["A"])&test_notisolated(ppi["B"])]
    t = ppi["score"].max()
    ppi_accepted = ppi.loc[ppi["score"]>=t]
    glist = [g for g in glist if (g not in isolated_genes)]
    components = get_weakly_connected(ppi_accepted, glist, index_col="A", column_col="B", score_col="score")
    if (not quiet):
        print("%d core genes in largest CC" % len([g for g in components[0] if (g in core_gene_set)]))
    isolated_genes = [g for c in components[1:] for g in c if (g not in components[0])]
    isolated_coregenes = [g for g in isolated_genes if (g in core_gene_set)]
    ## Add edges to/from isolated genes in decreasing score order
    ## until the PPI is connected and all genes in the connected PPI are present
    if (len(isolated_coregenes)>0):
        test_isolated = np.vectorize(lambda x : x in isolated_genes)
        ppi_rejected = ppi.loc[ppi["score"]<t]
        t_array = np.unique(ppi_rejected[["score"]].values)
        m, M = 0, len(t_array)
        while ((m!=M) or (M!=0) or (m!=len(t_array))):
            t = t_array[(M+m)//2]
            add_ppi = ppi_rejected.loc[ppi_rejected["score"]>=t]
            ppi_accepted_test = pd.concat((ppi_accepted, add_ppi), axis=0)
            components = get_weakly_connected(ppi_accepted_test, glist, index_col="A", column_col="B", score_col="score")
            if (not quiet):
                print("adding %d edges at t=%.6f (%d edges in total)" % (add_ppi.shape[0], t, ppi_accepted_test.shape[0]))
            isolated_genes = [g for c in components[1:] for g in c if (g not in components[0])]
            isolated_coregenes = [g for g in isolated_genes if (g in core_gene_set)]
            if (not quiet):
                print("%d components %s... (%d isolated genes of the core set)" % (len(components), str([len(c) for c in components[:3]]),len(isolated_coregenes)))
            if (len(isolated_coregenes)>0):
                M = (M+m)//2+1
            else:
                m = (M+m)//2
                break
        t = t_array[m]
    assert len([g for g in components[0] if (g in core_gene_set)])==len(main_component)
    return t

def merge_network_PPI(network, PPI, quiet=True):
    '''
        Merge two network while solving all inconsistencies (duplicates, paradoxes, etc.) in signs, directions, scores
        @param\tnetwork\tPandas DataFrame: rows/[interactions] x at least three columns "preferredName_A" (input node), "preferredName_B" (output node), "score" (edge score)
        @param\tPPI\tPandas DataFrame: rows/[interactions] x at least three columns "preferredName_A" (input node), "preferredName_B" (output node), "score" (edge score)
        @param\tquiet\tPython bool[default=None]: 
        @return\tfinal_network\tPandas DataFrame: rows/[interactions] x columns/[["preferredName_A", "preferredName_B", "sign", "directed", "score"]]
    '''
    edges_PPI = ["--".join(PPI.loc[x][["preferredName_A","preferredName_B"]]) for x in PPI.index]
    edges_network = ["--".join(network.loc[x][["preferredName_A","preferredName_B"]]) for x in network.index]
    if (not quiet):
        print("%d edges in network 1, %d in network 2, intersection=%d" % (len(edges_PPI),len(edges_network),len([x for ix, x in enumerate(network.index) if (edges_network[ix] in edges_PPI or "--".join([edges_network[ix].split("--")[-1], edges_network[ix].split("--")[0]]) in edges_PPI)])))
    add_indices = [x for ix, x in enumerate(network.index) if (edges_network[ix] not in edges_PPI and "--".join([edges_network[ix].split("--")[-1], edges_network[ix].split("--")[0]]) not in edges_PPI)]
    final_network = pd.concat((PPI, network.loc[add_indices]), axis=0)
    edges_final_network = ["--".join(final_network.loc[x][["preferredName_A","preferredName_B"]]) for x in final_network.index]
    scores = []
    for ix, x in enumerate(list(final_network["score"])):
        edge = edges_final_network[ix]
        inv_edge = "--".join([edge.split("--")[-1], edge.split("--")[0]])
        if (edge not in edges_network and inv_edge not in edges_network):
            scores.append(x)
        else:
            if (edge in edges_network):
                idx = edges_network.index(edge)
            elif (inv_edge in edges_network):
                idx = edges_network.index(inv_edge)
            x = network.iloc[idx,:]["score"]
            scores.append(x)
    final_network["score"] = scores
    return final_network

## https://workflows.omnipathdb.org/tissue-hpa.pdf
## https://workflows.omnipathdb.org/networks-r.html
def get_network_from_OmniPath(gene_list=None, disease_name=None, species="human", sources_int="omnipath", domains_int=None, types_int=None, min_curation_effort=-1, domains_annot='HPA_tissue', quiet=False):
    '''
        Retrieve a network from OmniPath
        @param\tgene_list\tPython character string[default=None]: List of genes to consider (or do not filter the interactions from Omnipath if =None)
        @param\tdisease_name\tPython character string[default=None]: Disease name (in letters) to consider
        @param\tspecies\tPython character string[default=None]: Species to consider (either "human", "mouse", or "rat")
        @param\tsources_int\tPython character string[default=None]: Which databases for interactions to consider (if =None, consider them all)
        @param\tdomains_int\tPython character string[default=None]:
        @param\ttypes_int\tPython character string[default=None]: Types of interactions, e.g., "post_translational", "transcriptional", "post_transcriptional", "mirna_transcriptional"
        @param\tmin_curation_effort\tPython integer[default=-1]: if positive, select edges based on that criteria (the higher, the better). Counts the unique database-citation pairs, i.e. how many times was an interaction described in a paper and mentioned in a database
        @param\tdomain_annot\tPython integer[default='HPA_tissue']:
        @param\tquiet\tPython bool[default=None]:  
        @return\tfinal_network\tPandas DataFrame: rows/[interactions] x columns/[["preferredName_A", "preferredName_B", "sign", "directed", "score"]]
                \tannot_wide\tPandas DataFrame: rows/[gene symbols] x columns/[annotations from the database @domains_annot]
    '''
    assert species in ['human', 'mouse', 'rat']
    assert types_int in [None, "post_translational", "transcriptional", "post_transcriptional", "mirna_transcriptional"]
    assert sources_int in list(op.interactions.AllInteractions.resources())
    assert domains_int in ["DOROTHEA", "KINASE_EXTRA", "LIGREC_EXTRA", "LNCRNA_MRNA", "MIRNA_TARGET","OMNIPATH",
        "PATHWAY_EXTRA", "SMALL_MOLECULE", "TF_MIRNA", "TF_REGULONS", "TF_TARGET"]
    assert domains_annot in [None]+list(op.requests.Annotations.resources())
    assert (gene_list is not None) or (disease_name is not None)
    params = {"organisms": species, "include": domains_int, "sources": sources_int, "fields": ['curation_effort'], #'references', 'sources', 'type'], 
              "genesymbols": True, "types": types_int, "directed": False, "signed": False} 
    with capture() as out:
        interactions = op.interactions.AllInteractions.get(**params).squeeze('columns')
    if (not quiet):
        print("Getting all interactions from OmniPath with the following parameters:%s..." % str(params))
    interactions = interactions[interactions["curation_effort"]>min_curation_effort]
    interactions["directed"] = interactions["consensus_direction"].astype(int)
    interactions["sign"] = interactions["consensus_stimulation"].astype(int)-interactions["consensus_inhibition"].astype(int)+2*(interactions["consensus_stimulation"].astype(int)*interactions["consensus_inhibition"].astype(int)+(1-interactions["consensus_stimulation"].astype(int))*(1-interactions["consensus_inhibition"].astype(int)))
    interactions = interactions[['source_genesymbol', 'target_genesymbol', 'directed', 'sign']]
    interactions.columns = ["preferredName_A","preferredName_B","directed","sign"]
    if (not quiet):
        print("Define a network-formatted dataframe from these interactions...")
    if (gene_list is not None):
        interactions = interactions[interactions.preferredName_A.isin(gene_list)&interactions.preferredName_B.isin(gene_list)]
    else:
        gene_list = list(set(list(interactions["preferredName_A"])+list(interactions["preferredName_B"])))
    if (not quiet):
        print("Obtain the gene list or filter interactions using the given gene list (%d genes)..." % len(gene_list))
    annot_ls = []
    chunks = np.array_split(gene_list, len(gene_list)//500+1)
    for i, gene_ls in enumerate(chunks):
        if (not quiet):
            print("%d/%d gene annotation chunks" % (i+1, len(chunks)))
        with capture() as out:
            annot_ls.append(op.requests.Annotations.get(proteins=gene_ls,resources=domains_annot).squeeze("columns"))
        annot = pd.concat(tuple(annot_ls), axis=0)
    annot = annot[["genesymbol", "label", "value"]]
    annot_wide = pd.pivot_table(annot, index='genesymbol', columns='label', values='value', aggfunc='first')
    return interactions, annot_wide

def remove_isolated(network, quiet=False):
    '''
        Remove all nodes which do not belong to the largest connected component from the network
        @param\tnetwork\tPandas DataFrame: rows/[interactions] x columns/[["preferredName_A", "preferredName_B", "sign", "directed", "score"]]
        @param\tquiet\tPython bool[default=None]:  
        @return\ttrimmed_network\tPandas DataFrame: rows/[interactions] x columns/[["preferredName_A", "preferredName_B", "sign", "directed", "score"]]
    '''
    glist = list(set(list(network["preferredName_A"])+list(network["preferredName_B"])))
    components = get_weakly_connected(network, glist, score_col="score")
    is_comp0 = np.vectorize(lambda x : x in components[0])
    if (not quiet):
        print("%d genes in total in the network (%d genes in the largest connected component)" % (len(glist), len(components[0])))
    network_trimmed = network.loc[is_comp0(network["preferredName_A"])&is_comp0(network["preferredName_B"])]
    if (not quiet):
        print("%d edges in trimmed network (down from %d edges)" % (network_trimmed.shape[0], network.shape[0]))
    return network_trimmed.sort_values(by="score", ascending=False)

def aggregate_networks(file_folder, gene_list, taxon_id, min_score, network_type, app_name, version_net="11.5", version_act="11.0", quiet=0):
    '''
        This function performs the following pipeline to build a prior knowledge network based on a subset of genes
	- Retrieve protein actions and predicted PPIs from STRING
	- Merge the two networks while solving all inconsistencies (duplicates, paradoxes, etc.) in signs, directions, scores
	- Determine the greatest threshold on the edge score which allows all of the core gene set to be connected (binary search)
	- Trim out edges which scores are below the threshold, and remove all isolated nodes
        @param\tfile_folder\tPython character string: relative path where to store files
        @param\tgene_list\tPython character string list: list of core gene symbols to preserve in the network
        @param\ttaxon_id\tPython integer: NCBI taxonomy ID
        @param\tmin_score\tPython integer: minimum score on edges retrieved from the STRING database
        @param\tapp_name\tPython character string: Identifier for STRING requests
        @param\tversion_net\tPython character string[default="11.5"]: Number of version for interaction data in the STRING database. To avoid compatibility issues, it is strongly advised not to change this parameter
        @param\tversion_act\tPython character string[default="11.0"]: Number of version for protein action data in the STRING database. To avoid compatibility issues, it is strongly advised not to change this parameter
        @param\tquiet\tPython bool[default=None]: 
        @return\tfinal_network\tPandas DataFrame: rows/[interactions] x columns/[["preferredName_A", "preferredName_B", "sign", "directed", "score"]]
    '''
    network = get_network_from_STRING(gene_list, taxon_id, min_score=min_score, network_type=network_type, add_nodes=0, app_name=app_name, version=version_net, quiet=quiet)
    PPI = get_interactions_from_STRING(gene_list, taxon_id, min_score=min_score, strict=False, version=version_act, app_name=app_name, file_folder=file_folder)
    final_network = merge_network_PPI(network, PPI)
    genes = list(set(list(final_network["preferredName_A"])+list(final_network["preferredName_B"])))
    threshold = determine_edge_threshold(final_network, genes)
    final_network_thres = final_network.loc[final_network["score"]>=threshold]
    ## remove redundant edges
    final_network_trimmed_thres = final_network_thres.sort_values(by="score", ascending=False).drop_duplicates(keep="first")
    ## remove isolated nodes
    final_network_trimmed_thresf = remove_isolated(final_network_trimmed_thres)
    return final_network_trimmed_thresf.sort_values(by="score", ascending=False)


# coding: utf-8

import bonesis
import mpbn
import numpy as np
import pandas as pd
from itertools import product
from functools import reduce
import sys
from zipfile import ZipFile
import os
from glob import glob
from tqdm import tqdm
from subprocess import call as sbcall
from copy import deepcopy
from sklearn.linear_model import LogisticRegression as Logit

from NORDic.UTILS.utils_state import quantile_normalize

def desirability(x, f_weight_di, A=0, B=1):
    '''
        Harrington's desirability function, used by [1]
        Convert a list of functions to maximize into a single scalar function to maximize with values in [@A,@B]
        @param\tx\tPoint
        @param\tf_weight_di\tPython dictionary: function with arguments as the same type as x, and associated weight
        @param\tA\tPython float[default=0]: lower bound of the function interval
        @param\tB\tPython float[default=1]: upper bound of the function interval
        @return\tdes(x)\tvalue of the desirability function at point x
        [1] http://ceur-ws.org/Vol-2488/paper17.pdf
        https://cran.r-project.org/web/packages/desirability/vignettes/desirability.pdf
    '''
    values = [np.exp(-np.exp(-((B-A)+weight/(B-A)*funct(x)))) for funct, weight in f_weight_di.items()]
    return float(np.exp(np.mean(np.log(values))))

def DS(influences):
    '''
        Computes the number of edges over the maximum number of possible connections between the nodes in the network
        @param\tinfluences\tPandas DataFrame: rows/[genes] x columns/[genes]
        @return\tDS\tPython float: network density
    '''
    n = influences.shape[0]
    return 2*np.sum(np.sum(np.triu(np.abs(influences.values)), axis=1), axis=0)/float(n*(n-1))

def CL(influences):
    '''
        Computes the average of node-wise clustering coefficients
        The clustering coefficient of a node is the ratio of the degree of the considered node and the maximum possible number of connections such that this node and its current neighbors form a clique
        @param\tinfluences\tPandas DataFrame: rows/[genes] x columns/[genes]
        @return\tCL\tPython float: network clustering coefficient
    '''
    n = influences.shape[0]
    network_outdegree = np.sum(np.abs(influences.values), axis=1)
    connect = np.sum(np.triu((np.abs(influences)+np.abs(influences).T).values),axis=1)
    max_connect = 0.5*np.multiply(network_outdegree,network_outdegree-1)
    max_connect[max_connect<=0] = 0.5
    numCL = np.sum(np.divide(connect, max_connect),axis=0)
    return numCL/n

def Centr(influences):
    '''
        Computes the network centralization, which is correlated with the similarity of the network to a graph with a star topology
        @param\tinfluences\tPandas DataFrame: rows/[genes] x columns/[genes]
        @return\tCentr\tPython float: network centralization
    '''
    n = influences.shape[0]
    network_outdegree = np.sum(np.abs(influences.values),axis=1)
    k_max = np.max(network_outdegree)
    return n/(n-2)*(k_max/(n-1)-DS(influences))

def GT(influences):
    '''
        Computes the network heterogeneity, which quantifies the non-uniformity of the node degrees across the network
        @param\tinfluences\tPandas DataFrame: rows/[genes] x columns/[genes]
        @return\tGT\tPython float: network heterogeneity
    '''
    network_outdegree = np.sum(np.abs(influences.values), axis=1)
    k_mean = np.mean(network_outdegree)
    return np.sqrt(np.var(network_outdegree))/max(k_mean,1)

def general_topological_parameter(influences, weights):
    '''
        Computes the general topological parameter (GTP) associated with the input network
        @param\tinfluences\tPandas DataFrame: rows/[genes] x columns/[genes]
        @param\tweights\tPython dictionary of (Python character string x Python float): all keys must be in ["DS","CL","Centr","GT"]
        @return\tscore\tPython float: return the score using the Harrington's desirability function
    '''
    assert all([w in ["DS","CL","Centr","GT"] for w in weights])
    missing_index = [x for x in influences.index if (x not in influences.columns)]
    missing_cols = [x for x in influences.columns if (x not in influences.index)]
    if (len(missing_index)>0):
        influences = pd.concat((influences, pd.DataFrame(0, index=missing_index, columns=influences.columns)), axis=0)
    if (len(missing_cols)>0):
        influences = pd.concat((influences, pd.DataFrame(0, index=list(influences.index)+missing_index, columns=missing_cols)), axis=1)
    return desirability(influences, {DS: weights.get("DS",0), CL: weights.get("CL",0), Centr: weights.get("Centr",0), GT: weights.get("GT", 0)})

def get_weakly_connected(network_df, gene_list, index_col="preferredName_A", column_col="preferredName_B", score_col="sscore"):
    '''
        Depth-first search (DFS) on undirected network
        @param\tnetwork_df\tPandas DataFrame: rows/[index] x columns/[["Input","Output"]]
        @param\tgene_list\tPython character string list: list of genes (needed to take into account isolated genes in the network)
        @param\tindex_col\tPython character string[default="preferredName_A"]: column in network_df (input gene)
        @param\tcolumn_col\tPython character string[default="preferredName_B"]: column in network_df (output gene)
        @param\tscore_col\tPython character string[default="sscore"]: column in network_df (edge weight)
        @return\tcomponents\tType of @network_df.loc[network_df.index[0]]["Input"] Python list of Python list: list of weakly connected components in the network, ordered by decreasing size
    '''
    adjacency = network_df.pivot_table(index=index_col, columns=column_col, values=score_col, aggfunc="mean")
    ## Undirected adjacency matrix
    adjacency[~np.isnan(adjacency)] = 1
    adjacency = adjacency.fillna(0)
    missing_index = [g for g in gene_list if (g not in adjacency.index)]
    if (len(missing_index)>0):
        missing_index_df = pd.DataFrame(0, index=missing_index, columns=adjacency.columns)
        adjacency = pd.concat((adjacency, missing_index_df), axis=0)
    missing_columns = [g for g in gene_list if (g not in adjacency.columns)]
    if (len(missing_columns)>0):
        missing_columns_df = pd.DataFrame(0, index=adjacency.index, columns=missing_columns)
        adjacency = pd.concat((adjacency, missing_columns_df), axis=1)
    assert adjacency.shape[0]==adjacency.shape[1]==len(gene_list)
    N = len(gene_list)
    components = []
    to_visit = [0]
    visited = [False]*N
    while (not all(visited)):
        component = []
        while (True):
            node = to_visit.pop()
            visited[node] = True
            if (node not in component):
                component.append(node)
            children = [list(adjacency.index).index(list(adjacency.columns)[i]) for i in np.argwhere(adjacency.values[node, :] != 0).flatten().tolist()]
            to_visit = [child for child in children if (not visited[child])]+to_visit
            if (len(to_visit)==0):
                break
        components.append(component)
        to_visit = [np.argmin(visited)]
    components = list(sorted(components, key=lambda c : len(c), reverse=True))
    components = [[list(adjacency.index)[n] for n in c] for c in components]
    return components

def get_genes_interactions_from_PPI(ppi, connected=False, score=0, filtering=True, quiet=False):
    '''
        Filtering edges to decrease computational cost while preserving network connectivity (if needed)
        @param\tppi\tPandas DataFrame: rows/[index] x columns[{"preferredName_A", "preferredName_B", "sign", "directed", "score"]]; sign in {-1,1,2}, directed in {0,1}, score in [0,1]
        @param\tconnected\tPython bool[default=True]: if set to True, preserve/enforce connectivity on the final network
        @param\tscore\tPython float[default=0]: Lower bound on the edge-associated score
        @param\tfiltering\tPython bool[default=True]: Whether to filter out edges by a correlation threshold
        @param\tquiet\tPython bool[default=False]
        @return\tppi_accepted\tPandas DataFrame: rows/[index] x columns/[["Input", "Output"]]
    '''
    assert all([x in [-1,1,2] for x in list(ppi["sign"])])
    assert all([x in [0,1] for x in list(ppi["directed"])])
    assert all([x <= 1 and x >= 0 for x in list(ppi["score"])])
    cols = ['preferredName_A', 'preferredName_B', 'sign', 'directed', 'score']
    assert all([col in cols for col in ppi.columns])
    assert ppi.shape[1]==len(cols)
    ppi = ppi[['preferredName_B', 'preferredName_A', 'sign', 'directed', 'score']]
    ## Remove duplicate edges (preserve the duplicates with highest score)
    ppi.index = ["--".join(x) for x in zip(ppi["preferredName_A"], ppi["preferredName_B"])]
    ppi = ppi.sort_values(by="score", ascending=False)
    ppi = ppi.loc[~ppi.index.duplicated(keep="first")]
    ppi = ppi[["preferredName_A","preferredName_B", 'sign', 'directed', 'score']]
    ppi.index = range(ppi.shape[0])
    Ntotal_edges = ppi.shape[0]
    ## 1. Double edges depending on whether they are marked as "directed"
    undirected_edges = ppi.loc[ppi["directed"]==0]
    Nundirected_edges = undirected_edges.shape[0]
    undirected_edges.columns = ['preferredName_B', 'preferredName_A', 'sign', 'directed', 'score']
    undirected_edges = undirected_edges[['preferredName_A', 'preferredName_B', 'sign', 'directed', 'score']]
    assert all([c==['preferredName_A', 'preferredName_B', 'sign', 'directed', 'score'][i] for i, c in enumerate(undirected_edges.columns)])
    assert undirected_edges.loc[undirected_edges.index[0]]['preferredName_B']==ppi.loc[ppi["directed"]==0].loc[ppi.loc[ppi["directed"]==0].index[0]]['preferredName_A']
    ppi = pd.concat((ppi, undirected_edges),axis=0)[["preferredName_A","preferredName_B","sign","score"]]
    ppi.index = range(ppi.shape[0])
    assert ppi.shape[0]==Ntotal_edges+Nundirected_edges
    ## 2. Double edges depending on whether they have a specific sign
    nonmonotonic_edges = ppi.loc[ppi["sign"]==2]
    Nnonmonotonic_edges = nonmonotonic_edges.shape[0]
    ppi_sign_sum = sum([abs(x) for x in list(ppi["sign"])])
    nonmonotonic_edges = pd.concat((nonmonotonic_edges[[c for c in nonmonotonic_edges.columns if (c!="sign")]], pd.DataFrame([[-1]]*nonmonotonic_edges.shape[0], index=nonmonotonic_edges.index, columns=["sign"])), axis=1)
    nonmonotonic_edges = nonmonotonic_edges[["preferredName_A","preferredName_B","sign","score"]]
    ppi = pd.concat((ppi, nonmonotonic_edges), axis=0)
    ppi["sign"] = [1 if (s==2) else s for s in list(ppi["sign"])]
    ppi.index = range(ppi.shape[0])
    assert ppi.shape[0]==Ntotal_edges+Nundirected_edges+Nnonmonotonic_edges
    assert ppi["sign"].abs().sum().sum()==ppi_sign_sum
    ## 3. Concatenate columns "score" and "sign" as score in [0,1]
    ppi["sscore"] = np.multiply(ppi["score"], ppi["sign"])
    ppi = ppi[["preferredName_A","preferredName_B","sscore"]]
    if (connected):
        components = get_weakly_connected(ppi, list(set(list(ppi["preferredName_A"])+list(ppi["preferredName_B"]))))
        main_component = components[0]
        ## 5. Remove genes which are isolated in the full PPI
        isolated_genes = [g for c in components[1:] for g in c]
        if (len(isolated_genes)>0):
            test_isolated = np.vectorize(lambda x : x in isolated_genes)
            ppi = ppi.loc[~test_isolated(ppi["preferredName_A"])&~test_isolated(ppi["preferredName_B"])]
        t = min(score, ppi["sscore"].abs().max()) if (filtering) else ppi["sscore"].abs().min()
        ppi_accepted = ppi.loc[ppi["sscore"].abs()>=t]
        components = get_weakly_connected(ppi_accepted, main_component)
        isolated_genes = [g for c in components[1:] for g in c if (g not in components[0])]
        assert len(isolated_genes)+len(components[0])==len(main_component)
        ## 6. Add edges to/from isolated genes in decreasing (absolute) score order until the PPI is connected and all genes in the connected PPI are present
        if (len(isolated_genes)>0):
            ppi_rejected = ppi.loc[ppi["sscore"].abs()<t]
            test_isolated = np.vectorize(lambda x : x in isolated_genes)
            while ((len(isolated_genes)>0) and (ppi_rejected.shape[0]>0)):
                ppi_isolated = ppi_rejected.loc[(test_isolated(ppi_rejected["preferredName_A"])&~test_isolated(ppi_rejected["preferredName_B"]))|(test_isolated(ppi_rejected["preferredName_B"])&~test_isolated(ppi_rejected["preferredName_A"]))]
                t = np.max(ppi_isolated["sscore"].abs())
                add_ppi_isolated = ppi_isolated.loc[ppi_isolated["sscore"].abs()>=t]
                ppi_accepted = pd.concat((ppi_accepted, add_ppi_isolated), axis=0)
                components = get_weakly_connected(ppi_accepted, main_component)
                if (not quiet):
                    print("adding %d edges at t=%f (%d edges in total)" % (add_ppi_isolated.shape[0], t, ppi_accepted.shape[0]))
                isolated_genes = [g for c in components[1:] for g in c if (g not in components[0])]
                if (not quiet):
                    print("Components %s... (%d isolated genes)" % (str([len(c) for c in components[:3]]),len(isolated_genes)))
                assert len(isolated_genes)+len(components[0])==len(main_component)
                test_isolated = np.vectorize(lambda x : x in isolated_genes)
                ppi_rejected = ppi_rejected.loc[(ppi_rejected["sscore"].abs()<t)|((test_isolated(ppi_rejected["preferredName_A"])|(test_isolated(ppi_rejected["preferredName_B"]))))]
        assert len(components)==1 and len(components[0])==len(main_component)
    else:
        t = min(score, ppi["sscore"].abs().max()) if (filtering) else ppi["sscore"].abs().min()
        ppi_accepted = ppi.loc[ppi["sscore"].abs()>=t]
    ppi_accepted.columns = ["Input","Output","SSign"]
    return ppi_accepted

def build_influences(network_df, tau, beta=1, cor_method="pearson", expr_df=None, accept_nonRNA=False, quiet=False):
    '''
        Filters out (and signs of unsigned) edges based on gene expression 
        @param\tnetwork_df\tPandas DataFrame: rows/[index] x columns/[["Input", "Output", "SSign"]] interactions
        @param\ttau\tPython float: threshold on genepairwise expression correlation
        @param\tbeta\tPython integer[default=1]: power applied to the adjacency matrix
        @param\tcor_method\tPython character string[default="pearson"]: type of correlation
        @param\texpr_df\tPandas DataFrame[default=None]: rows/[genes] x columns/[samples] gene expression data
        @param\taccept_nonRNA\tPython bool[default=False]: if set to False, ignores gene names which are not present in expr_df
        @param\tquiet\tPython bool[default=False]
        @return\tinfluences\tPandas DataFrame: rows/[genes] x columns/[genes] signed adjacency matrix with only interactions s.t. corr^beta>=tau
    '''
    assert network_df.shape[1]==3 and all([c in ["Input","Output", "SSign"] for c in network_df.columns])
    network = network_df.pivot_table(index="Input", columns="Output", values="SSign", aggfunc="mean")
    missing_index = [g for g in network.columns if (g not in network.index)]
    if (len(missing_index)>0):
        missing_index_df = pd.DataFrame([], index=missing_index, columns=network.columns)
        network = pd.concat((network, missing_index_df), axis=0)
    missing_columns = [g for g in network.index if (g not in network.columns)]
    if (len(missing_columns)>0):
        missing_columns_df = pd.DataFrame([], index=network.index, columns=missing_columns)
        network = pd.concat((network, missing_columns_df), axis=1)
    assert network.shape[0]==network.shape[1]
    network = network[network.index]
    network = network.fillna(-666) #missing value
    network[network==0] = 2 #means that the edge is present both as activatory and inhibitory
    network[network==-666] = 0
    network[network<0] = -1 #only inhibitory
    network[(network>0)&(network<2)] = 1 #only activatory
    network = network.astype(int)
    network_unsigned, network_signed = [deepcopy(network) for _ in range(2)] 
    network_unsigned[network_unsigned!=2] = 0
    network_unsigned = network_unsigned/2 #1 if unsigned edge exists, 0 otherwise
    network_signed[network_signed==2] = 0 #1 if signed activatory edge exists, -1 if signed inhibitory edge exists, 0 otherwise
    # Correlation matrix
    if (expr_df is not None):
        df = quantile_normalize(expr_df)
        df = df.loc[[g for g in list(network.index) if (g in df.index)]]
        coexpr = np.power(df.T.corr(method=cor_method), beta)
        coexpr[coexpr.abs()<tau] = 0
        if (accept_nonRNA):
            missing_genes = [g for g in list(network.index) if (g not in df.index)]
            for g in missing_genes:
                coexpr[g] = 1
                coexpr.loc[g] = 1
            coexpr = coexpr.loc[list(network.index)][list(network.index)]
    else:
        coexpr = pd.DataFrame(np.ones(network.shape), index=list(network.index), columns=list(network.index)) #default: positive interactions
    assert all([coexpr.shape[i]==s for i,s in enumerate(network_unsigned.shape)])
    assert all([coexpr.index[i]==s for i,s in enumerate(network_unsigned.index)])
    assert all([coexpr.columns[i]==s for i,s in enumerate(network_unsigned.columns)])
    net_mat = np.multiply((-1)**(coexpr.values<0).astype(int), network_unsigned.values).astype(float)#equal to 1 if strong correlation and activatory, -1 if strong correlation and inhibitory, 0 otherwise
    influences = pd.DataFrame(net_mat+network_signed.values, index=network.index, columns=network.columns)
    assert not pd.isnull(influences).any().any()
    return influences

def create_grn(influences, exact=False, max_maxclause=3, quiet=False):
    '''
        Create a BoneSiS InfluenceGraph
        @param\tinfluences\tPandas DataFrame: rows/[genes] x columns/[genes] of interactions, values in {-1,1,0} -1:negative,1:positive,0:absent
        @param\texact\tPython bool[default=False]: should all interactions be preserved?
        @param\tmax_maxclause\tPython integer[default=3]: upper bound on the number of clauses in DNF form
        @param\tquiet\tPython bool[default=False]
        @return\tgrn\tBoneSiS InfluenceGraph class object
    '''
    assert all([n in [-1,1,0] for n in np.unique(influences.values)])
    maxclause = min(get_maxdegree(influences, quiet=quiet), max_maxclause)
    influences_list = influences.melt(ignore_index=False)
    influences_list["id"] = influences_list.index
    influences_list = influences_list[influences_list["value"]!=0]
    influences_list = [(x,y,dict(sign=s)) for [y,s,x] in influences_list.values.tolist()]
    if (not quiet):
        print("<BONESIS> %d interactions (maximum # of clauses = %d)" % (len(influences_list), maxclause))
    grn = bonesis.InfluenceGraph(graph=influences_list, maxclause=maxclause, exact=exact)
    return grn

def get_maxdegree(influences, activatory=True, quiet=False):
    '''
        Computes the maximum ingoing degree (or the maximum number of potential activatory regulators) in a graph
        @param\tinfluences\tPandas DataFrame: rows/[genes] x columns/[genes] of interactions: -1:negative,1:positive,0:absent
        @param\tactivatory\tPython bool[default=True]: computes the maximum number of potential activatory regulators instead
        @param\tquiet\tPython bool[default=False]
        @return\tmaxindegree\tPython integer
    '''
    assert all([n in [-1,0,1] for n in np.unique(influences.values)])
    if (not activatory):
        maxindegree=int(np.max((influences!=0).astype(float).sum(axis=0)))
    else:
        maxindegree=int(np.max((influences>0).astype(float).sum(axis=0)))
    if (not quiet):
        print("<UTILS_GRN> Maximum "+("ingoing degree" if (not activatory) else "possible #activators")+"=%d" % maxindegree)
    return maxindegree

def build_observations(grn, signatures, quiet=False):
    '''
        Implement experimental constraints from perturbation experiments in signatures
        Experimental constraints are of the form...
        @param\tgrn\tInfluenceGraph (from BoneSiS): contains topological constraints
        @param\tsignatures\tPandas DataFrame: rows/[genes] x columns/[experiment IDs]. Experiment IDs is of the form "<pert. gene>_<pert. type>_<...>_<cell line>" (treated) or "initial_<cell line>" (control)
        @param\tquiet\tPython bool[default=False]
        @return\tBO\tBoNesis object (from BoneSiS)
    '''
    data_exps = {}
    if (len(signatures)==0):
        BO = bonesis.BoNesis(grn, data_exps)
        return BO
    signatures = signatures.loc[[g for g in signatures.index if (g in grn.nodes)]]
    ## 1. Add signatures of experimental states
    exps = [x for x in signatures.columns if ("initial" not in x)]
    exps_ids = range(len(exps))
    ## 1a. Instantiate experiments
    ## zero state to ignore
    #data_exps.update({"zero": {g: 0 for g in grn.nodes}})
    ## 1b. For each experiment (1 initial state, 1 final state, 1 perturbation)
    for exp_nb in exps_ids:
        cell = exps[exp_nb].split("_")[-1]
        cols = ["Exp%d_"%(exp_nb+1)+x for x in ["init", "final"]]
        data_df = signatures[["initial_"+cell, exps[exp_nb-1]]]

        ## Compatible with perturbation experiment
        data_df = data_df.T
        pert, sign = exps[exp_nb].split("_")[0], 0 if (exps[exp_nb].split("_")[1]=="KD") else 1
        data_df[pert] = sign 
        data_df = data_df.T

        data_df.columns = cols
        for col in cols:
            data_exps.update(data_df[[col]].dropna().astype(int).to_dict())
    if (not quiet):
        print_exps = pd.DataFrame.from_dict(data_exps, orient="index").fillna(-1).astype(int)
        print_exps[print_exps==-1] = ""
        print("\n<UTILS_GRN> %d experiments\n%s" % (len(exps_ids), str(print_exps)))
    BO = bonesis.BoNesis(grn, data_exps)
    ## 2. Instantiate reachability & fixed point constraints
    for exp_nb in exps_ids:
        exp = exps[exp_nb]
        pert, sign = exp.split("_")[:2]
        cell = exp.split("_")[-1]
        sign = int(sign!="KD")
        with BO.mutant({pert: sign}) as m:
            final_FP = m.fixed(~m.obs("Exp%d_final" % (exp_nb+1)))
            ## 2a. There exists a trajectory : init -> trapspace
            ~m.obs("Exp%d_init" % (exp_nb+1)) >= final_FP #~m.obs("Exp%d_final" % (exp_nb+1)) ##
            ## MEMO: Universal reachable fixed points: ~m.obs("Exp%d_init" % (exp_nb+1)) >> "fixpoints" ^ {m.obs("Exp%d_final" % (exp_nb+1))}
        ## 2b. No trivial trajectory from zero state : zero -> trapspace(unless the experiment actually involves it)
        #state_di = data_exps["Exp%d_init" % (exp_nb+1)]
        #state_di.update({pert: sign})
        #if (all([state_di.get(g, 1)==0 for g in grn.nodes])):
        #    continue
        #~BO.obs("zero") / ~BO.obs("Exp%d_final" % (exp_nb+1))
    ## MEMO: Fixed point: BO.fixed(~BO.obs("attractor_%d" % (state_id+1)))
    ## MEMO: Trap space: BO.fixed(BO.obs("attractor_%d" % (state_id+1)))
    return BO

def solution2influences(solution):
    '''
        Converts a solution object into a influences object
        @param\tsolution\tPandas Series: rows/[genes]
        @return\tinfluences\tPandas DataFrame: rows/[genes] x columns/[genes] contains values {-1,1,0,2} -1: negative, 1: positive, 0: absent, 2: non monotonic
    '''
    gene_list = list(solution.index)
    N = len(gene_list)
    infl_mat = np.zeros((N,N))
    grfs = get_grfs_from_solution(solution)
    for gene in grfs:
        ig = gene_list.index(gene)
        for regulator in grfs[gene]:
            if (regulator in gene_list):
                ir = gene_list.index(regulator)
                infl_mat[ir,ig] = grfs[gene][regulator]
            else:
                infl_mat = np.concatenate((infl_mat, np.zeros((1,N))), axis=0)
                infl_mat = np.concatenate((infl_mat, np.zeros((N+1,1))), axis=1)
                infl_mat[N,ig] = grfs[gene][regulator]
                gene_list.append(regulator)
                N += 1
    influences = pd.DataFrame(infl_mat, index=gene_list, columns=gene_list)
    return influences

def zip2df(fname):
    '''
        Extract solutions in ZIP file as DataFrames
    '''
    with ZipFile(fname, "r") as zip:
        zip.extractall()
    grn_fnames = glob("*.bnet")
    grns = []
    for grn_fname in grn_fnames:
        grn = load_grn(grn_fname)
        grns.append(grn)
    solutions = pd.DataFrame(grns)
    sbcall("rm -f *.bnet", shell=True)
    return solutions

def load_grn(fname):
    '''
        Loads GRN as MPBN class element
    '''
    BN = mpbn.MPBooleanNetwork(fname)
    sep=" <- " if (" <- " in str(BN)) else ", "
    solution = pd.Series([x.split(sep)[-1] for x in str(BN).split("\n")], index=[x.split(sep)[0] for x in str(BN).split("\n")])
    return solution

def get_minimal_edges(R, maximal=False):
    '''
       Return one of the solutions with the smallest (or greatest) number of edges
       @param\tR\tPandas DataFrame: rows/[genes] x columns/[solution IDs]
       @param\tconnected\tPython bool[default=False]: if set to True, return the CONNECTED solution which satisfies those constraints
       @param\tmaximal\tPython bool[default=False]: if set to True, return the solution with the greatest number of edges
       @return\tsolution, nedges\tPython integer x Python integer: solution and corresponding number of edges
    '''
    nedges_list = [int(solution2influences(R[col]).abs().sum().sum()) for col in R.columns]
    solution_id = (np.argmax if (maximal) else np.argmin)(nedges_list)
    return R[R.columns[solution_id]], nedges_list[solution_id]

def get_grfs_from_solution(solution):
    '''
        Retrieve all gene regulatory functions (GRFs) from a given solution
        @param\tsolution\tPandas Series: rows/[genes] 
        @return\tgrfs\tPython dictionary: {gene: {regulator: sign, ...}, ...} where sign in {-1,1} -1: inhibitor, 1: activator
    '''
    sol_dict, grfs = solution.to_dict(), {}
    for gene in sol_dict:
        grf = "".join("".join(" ".join(" ".join(str(sol_dict[gene]).split(", ")[-1].split("|")).split("&")).split("(")).split(")"))
        if (len(str(grf))==0):
            continue
        regulators = {}
        if (str(grf) not in ["0", "1"]):
            reg_names = list(set(grf.split(" ")))
            regulators_df = pd.DataFrame([[(-1)**int(regulator[0]=="!")] for regulator in reg_names], index=[regulator[int(regulator[0]=="!"):] for regulator in reg_names], columns=["regulators"])
            regulators_df = regulators_df.groupby(level=0).sum()
            regulators_df[regulators_df==0] = 2
            regulators = regulators_df.to_dict()["regulators"]
        grfs.setdefault(gene, regulators)
    return grfs

def save_grn(solution, fname, sep=", ", quiet=False, max_show=5, write=True):
    '''
        Write and/or print .bnet file
        @param\tsolution\tPandas Series: rows/[genes] contains gene regulatory functions (GRF)
        @param\tfname\tPython character string: where to write the file (w/o .bnet extension)
        @param\tsep\tPython character string: what separates regulators from regulated genes
        @param\tquiet\tPython bool[default=False]
        @param\tmax_show\tPython integer[default=5]: maximum number of printed GRFs
        @param\twrite\tPython bool[default=True]: if set to True, write to a .bnet file
        @return\tNone\t
    '''
    sol = solution.to_dict()
    print_sol = ["".join([str(x) for x in [gene, sep, sol[gene]]]) for gene in sol if (len(gene)>0)]
    if (not quiet):
        print("\n"+("\n".join(print_sol[:max_show]+["..." if (len(print_sol)>max_show) else ""])))
    print_sol = "\n".join(print_sol)
    if (write):
        with open(fname+".bnet", "w+") as f:
            f.write(print_sol)
    return None

def save_solutions(bnetworks, fname, limit):
    '''
        Enumerate and save solutions
        @param\tbnetworks\tOutput of the inference
        @param\tfname\t
        @param\tlimit\tPython integer: maximum number of solutions to enumerate
        @return\tn\tPython integer: number of enumerated solutions
    '''
    with ZipFile(fname, "w") as bundle:
        n = 0
        for i, bn in enumerate(bnetworks):
            with bundle.open(f"bn{i}.bnet", "w") as fp:
                fp.write(bn.source().encode())
            n += 1
            if n == limit:
                break
    return n

def infer_network(BO, njobs=1, fname="solutions", use_diverse=True, limit=50, niterations=1):
    '''
        Infer solutions matching topological & experimental constraints
        @param\tBO\t Bonesis object (from BoneSiS): contains topological & experimental constraints
        @param\tfname\tPython character string[default="solutions"]: path to solution files
        @param\tuse_diverse\tPython bool[default=True]: use the "diverse" procedure in BoneSiS
        @param\tlimit\tPython integer[default=50]: maximum number of solutions to generate per interation
        @param\tniterations\tPython integer[default=1]: maximum number of iterations
        @return list of # solutions per iteration
    '''
    bonesis.settings["parallel"] = njobs
    #bonesis.settings["solutions"] = "subset-minimal"
    if (use_diverse):
        infer = lambda bo : bo.diverse_boolean_networks
    else:
        infer = lambda bo : bo.boolean_networks
    param_di = {"skip_supersets": True}
    if (str(limit)!="None"):
        param_di.setdefault("limit",limit)
    bnetworks = infer(BO)(**param_di)
    nsolutions = []
    for niter in tqdm(range(niterations)):
        if (not os.path.exists(fname+"_"+str(niter+1)+".zip")):
            nsolutions.append(save_solutions(bnetworks, fname+"_"+str(niter+1)+".zip", limit))
    return nsolutions

def get_genes_downstream(network_fname, gene, n=-1):
    '''
        Get the list of genes downstream of a gene in a network
        @param\tnetwork_fname\tPython character string: path to the .BNET file associated with the network
        @param\tgene\tPython character string: gene name in the network
        @param\tn\tPython integer[default=-1]: number of recursions (if<0, recursively get all downstream genes)
        @return\tlst_downstream\tPython character string list: list of nodes downstream of @gene
    '''   
    with open(network_fname, "r") as f:
        grf_list = f.read().split("\n")
    grfs = dict([x.split(", ") for x in grf_list if (len(x)>0)]) 
    set_downstream = set([])
    if (gene not in grfs):
        return []
    count_n = 0
    while ((n<0) or (count_n<n)):
        l = len(set_downstream)
        for g in grfs:
            if (any(x in grfs[g] for x in set_downstream.union({gene}))):
                set_downstream = set_downstream.union({g})
        if (l==len(set_downstream)):
            break
        count_n += 1
    return list(set_downstream)

def get_genes_most_variable(control_profiles, treated_profiles, p=0.8):
    '''
        Get the list of genes which contribute most to the variation between two conditions (in the @pth percentile of change)
        @param\tcontrol_profiles\tPandas DataFrame: rows/[genes] x columns/[samples] profiles from condition 1
        @param\ttreated_profiles\tPandas DataFrame: rows/[genes] x columns/[samples] profiles from condition 1
        @param\tp\tPython float: 100*p th percentile to consider 
        @return\tlst_genes\tPython character string list: list of nodes which contribute most to the variation between conditions
    '''  
    assert p>=0 and p<=1
    model = Logit(penalty="l1",solver="saga",fit_intercept=False,random_state=0,max_iter=1000,n_jobs=njobs)
    df = control_profiles.join(treated_profiles, how="inner").fillna(0.5)
    X = df.T.values
    y = np.ravel(np.array([0]*control_profiles.shape[1]+[1]*treated_profiles.shape[1]))
    model.fit(X, y)
    q = np.quantile(np.abs(model.coef_.flatten()), p)
    genes = {g: model.coef_[0,ig] for ig, g in enumerate(list(df.index)) if (abs(model.coef_[0,ig])>q)}
    return list(genes.keys())

def reconnect_network(network_fname):
    '''
        Write the network with all isolated nodes (no ingoing/outgoing edges) filtered out
        @param\tnetwork_fname\tPython character string: path to the .BNET associated with the network
        @return\tfname\tPython character string: path to the .BNET associated with the reconnected network
    '''  
    with open(network_fname, "r") as f:
        network = pd.DataFrame({"Solution": dict([["_".join(g.split("-")) for g in x.split(", ")] for x in f.read().split("\n") if (len(x)>0)])})
    influences = solution2influences(network["Solution"])
    assert influences.shape[0]==influences.shape[1]
    assert all([influences.index[i]==influences.columns[i] for i in range(influences.shape[0])])
    influences = influences.loc[(influences.abs().sum(axis=1)>0)&(influences.abs().sum(axis=0)>0)]
    gene_list = list(influences.index)
    network_connected = network.loc[gene_list]
    def get_all_genes(network):
        all_genes = list(network.index)
        for idx in network.index:
            grf = str(network.loc[idx]["Solution"])
            for symb in ["!","&","|","(",")"]:
                grf = " ".join(grf.split(symb))
            gene_lst = [g for g in grf.split(" ") if (g not in [" ", "0","1"] and len(g)>0)]
            all_genes += gene_lst
        return list(set(all_genes))
    full_lst = get_all_genes(network_connected)
    network_connected = network.loc[full_lst].to_dict()["Solution"]
    network_connected = "\n".join([", ".join(["_".join(g.split("-")) for g in [x,y]]) for x,y in network_connected.items()])
    network_fname_connected = network_fname.split(".bnet")[0]+"_connected.bnet"
    with open(network_fname_connected, "w") as f:
        f.write(network_connected) 
    return network_fname_connected



# coding:utf-8

## https://maayanlab.cloud/L1000CDS2/help/#cosine

import pandas as pd
import requests
import json
import numpy as np

from NORDic.UTILS.LINCS_utils import build_url, get_user_key, binarize_via_CD, download_lincs_files, create_restricted_drug_signatures
from NORDic.UTILS.utils_data import get_all_celllines

def get_ranking(CD):
    '''
        Retrieve ranking (50 first drugs) from L1000 CDS^2 search engine
        @param\tCD\tPandas DataFrame: rows/[genes] x column/[value] differential phenotype
        @return\tresuls\tPandas DataFrame: rows/[drug names] x column/["L1000 CDS2"] ranking 
        of the drugs according to their ability to reverse the phenotype
    '''
    ## # The app uses uppercase gene symbols. So it is crucial to perform this step
    CD.index = [g.upper() for g in CD.index]
    l1000cds2_url = 'https://maayanlab.cloud/L1000CDS2/query'
    ## https://www.mongodb.com/docs/manual/tutorial/install-mongodb-on-debian/
    ## https://maayanlab.cloud/public/L1000CDS_download/#genesMeta
    data = {"genes":list(CD.index),"vals":list(CD[CD.columns[0]])}
    config = {"aggravate":False,"searchMethod":"CD","share":False,"combination":False,"db-version":"latest","limit":100}
    headers = {'content-type': 'application/json'}
    results = {'topMeta': {}, 'uniqInput': []}
    metadata = [{"key":"Tag","value":"CD baseline test"}]
    payload = {"data":data,"config":config,"meta":metadata}
    r = requests.post(l1000cds2_url,data=json.dumps(payload),headers=headers,stream=True)
    assert r.status_code==200
    resCD = r.json()
    assert "err" not in resCD
    results["topMeta"].update({res["pert_desc"]: res["score"] for res in resCD["topMeta"]})# if (res["cell_id"] in ["NPC","SH-SY5Y"])}) 
    results["uniqInput"] = resCD["uniqInput"]
    return pd.DataFrame({"L1000 CDS2": results["topMeta"]}).sort_values(by="L1000 CDS2", ascending=False)

def drugname2pubchem(drug_names, lincs_args):
    '''
        Convert drug names into PubChem CIDs
        @param\tdrug_names\tPython character string list: list of drug names
        @param\tlincs_args\tPython dictionary: additional arguments for LINCS L1000 requests
        @return\tpubchem_cids\tPython dictionary: (keys=drug names, values=PubChem CIDs)
    '''
    endpoint = "perts"
    method = "filter"
    pubchem_cids = {}
    assert lincs_args and "credentials" in lincs_args
    user_key = get_user_key(lincs_args["credentials"])
    for ip, drug_name in enumerate(drug_names):
        params = {"where": {"pert_iname": drug_name.lower()}, "fields": ["pubchem_cid"]}
        request_url = build_url(endpoint, method, params=params, user_key=user_key)
        response = requests.get(request_url)
        assert response.status_code == 200
        data = json.loads(response.text)
        cid = int(data[0]["pubchem_cid"]) if (len(data)>0) else np.nan
        pubchem_cids.setdefault(drug_name, cid)
    return pubchem_cids

def pubchem2drugname(pubchem_cids, lincs_args):
    '''
        Convert drug names into PubChem CIDs
        @param\tpubchem_cids\tPython integer list: list of drug PubChem CIDs
        @param\tlincs_args\tPython dictionary: additional arguments for LINCS L1000 requests
        @return\tpert_inames\tPython dictionary: (keys=PubChem CIDs, values=drug names)
    '''
    endpoint = "perts"
    method = "filter"
    pert_inames = {}
    assert lincs_args and "credentials" in lincs_args
    user_key = get_user_key(lincs_args["credentials"])
    for ip, pubchem_cid in enumerate(pubchem_cids):
        params = {"where": {"pubchem_cid": pubchem_cid}, "fields": ["pert_iname"]}
        request_url = build_url(endpoint, method, params=params, user_key=user_key)
        response = requests.get(request_url)
        assert response.status_code == 200
        data = json.loads(response.text)
        drug_name = data[0]["pert_iname"] if (len(data)>0) else np.nan
        pert_inames.setdefault(pubchem_cid, drug_name)
    return pert_inames

def retrieve_drug_signature(pubchem_cid, cell_ids, gene_list, lincs_args, binarize, quiet=False):
    '''
        Retrieve control & treated samples from LINCS L1000 and compute the corresponding drug signature
        @param\tpubchem_cid\tPython integer: drug PubChem CID
        @param\tcell_ids\tPython character string list: list of candidate cell lines in LINCS L1000
        @param\tgene_list\tPython integer list: list of EntrezID genes
        @param\tlincs_args\tPython dictionary: additional arguments for LINCS L1000 requests
        @param\tbinarize\tPython bool: should the resulting signatures be binarized?
        @param\tquiet\tPython bool[default=False]
        @return\tsig\tPandas DataFrame: rows/[genes] x column/[drug PubChem]
    '''
    endpoint = "sigs"
    method = "filter"
    pert_iname = pubchem2drugname([pubchem_cid], lincs_args)[pubchem_cid]
    assert lincs_args and "credentials" in lincs_args
    user_key = get_user_key(lincs_args["credentials"])
    selection = lincs_args.get("selection", "distil_ss")
    data = []
    if (not quiet):
        print("* PubChemCID %d (name %s)" % (pubchem_cid, pert_iname))
    for cell_id in cell_ids:
        ## 1. Build request to get distil_id of relevant treated samples
        where = {"pert_type": "trt_cp", "pert_iname": pert_iname, "cell_id": cell_id}
        params = {
            "where": where,
            "fields": ["cell_id", "distil_id", "brew_prefix"]+[selection]
        }
        request_url = build_url(endpoint, method, params=params, user_key=user_key)
        response = requests.get(request_url)
        assert response.status_code == 200
        data += json.loads(response.text)
    if (len(data)==0):
        return None
    ## At least @nsigs replicates and maximize the criterion
    data_treated = [dt for dt in data if (len(dt["distil_id"])>lincs_args.get("nsigs", 2))]
    data_treated = data_treated[np.argmax([dt[selection] for dt in data_treated])]
    brew_prefix, cell_id, data_treated = data_treated["brew_prefix"], data_treated["cell_id"], data_treated["distil_id"]
    ## 2. Get corresponding control samples from the same plate
    where = {'pert_type': "ctl_vehicle", "cell_id": cell_id, "brew_prefix": brew_prefix}
    params = {"where": where,
        "fields": ["distil_id"]+[selection],
    }
    request_url = build_url(endpoint, method, params=params, user_key=user_key)
    response = requests.get(request_url)
    assert response.status_code == 200
    data = json.loads(response.text)
    if (len(data)==0):
        return None
    ## At least @nsigs replicates and maximize the criterion
    data_control = [dt for dt in data if (len(dt["distil_id"])>lincs_args.get("nsigs", 2))]
    data_control = data_control[np.argmax([dt[selection] for dt in data_control])]
    data_control = data_control["distil_id"]
    nsamples = len(data_treated+data_control)
    if (not quiet):
        print("%d samples" % nsamples)
    sigs = create_restricted_drug_signatures(data_treated+data_control, gene_list, which_lvl=[3], strict=True, path_to_lincs=lincs_args["path_to_lincs"])
    if (sigs is None):
        return None
    assert sigs.shape[1] == nsamples
    sig = binarize_via_CD(sigs, samples=[2]*len(data_treated)+[1]*len(data_control), binarize=int(binarize), nperm=10000)
    sig.columns = [pubchem_cid]
    return sig

def compute_drug_signatures_L1000(pubchem_cids, lincs_args, binarize=True, gene_list=None, chunksize=10):
    '''
        Get drug signatures from LINCS L1000
        @param\tpubchem_cids\tPython integer list: list of drug PubChem CIDs
        @param\tlincs_args\tPython dictionary: additional arguments for LINCS L1000 requests
        @param\tbinarize\tPython bool[default=True]: should the resulting signatures be binarized?
        @return\tsigs\tPandas DataFrame: rows/[genes] x columns/[drug names]
    '''
    assert lincs_args and "credentials" in lincs_args and "path_to_lincs" in lincs_args
    user_key = get_user_key(lincs_args["credentials"])
    path_to_lincs = lincs_args["path_to_lincs"]
    pert_inames = pubchem2drugname(pubchem_cids, lincs_args)
    gene_files, _, _, _ = download_lincs_files(path_to_lincs, which_lvl=[3])
    gene_df = pd.read_csv(path_to_lincs+gene_files[0], sep="\t", engine='python', index_col=0)
    gene_list, gene_name_list = list(gene_df.index) if (gene_list is None) else gene_list, list(gene_df["pr_gene_symbol"])
    sigs_list = [retrieve_drug_signature(pubchem_cid, get_all_celllines([pert_inames[pubchem_cid]], user_key) if (len(lincs_args.get("cell_lines", []))==0) else lincs_args["cell_lines"], gene_list, lincs_args, binarize) for pubchem_cid in pubchem_cids]
    sigs = sigs_list[0].join(sigs_list[1:], how="outer")
    sigs.index = [gene_df.loc[i]["pr_gene_symbol"] for i in sigs.index]
    return sigs


# coding:utf-8

import pandas as pd
import numpy as np

from joblib import Parallel, delayed
import mpbn
import mpbn.simulation as mpbn_sim
from tqdm import tqdm

from sklearn.metrics import pairwise_distances
from sklearn.metrics import accuracy_score as ACC
from sklearn.metrics import roc_auc_score as AUC
from sklearn.metrics import fbeta_score as Fbeta
from random import seed as rseed
from scipy.stats import ks_2samp, spearmanr, kendalltau

#################################################
## Baseline (L1000 CDS^2) [Duan et al., 2016]  ##
#################################################
def baseline(signatures, phenotype, is_binary=False):
    '''
        Compute the cosine scores between a set of signatures and a differential phenotype
        @param\tS\tPandas DataFrame: signatures rows/[genes] x columns/[drug names] [Treated || Control]
        @param\tP\tPandas DataFrame: differential phenotype rows/[genes] x column/[disease name] [Diseased || Healthy]
        @param\tis_binary\tPython bool[default=False]: if set to True, the signatures and phenotypes might be binary (that is, with values in {0,1})
        @return\tscores\tPython float dictionary: dictionary (keys=drug names, values=scores), the higher the score, the higher the repurposing power
    '''
    SC, PC = signatures.dropna(), phenotype.dropna()
    if (is_binary):
        SC[SC>0] = 1
        PC[PC>0] = 1
        SC[SC<0] = -1
        PC[PC<0] = -1
        SC[pd.isnull(SC)] = 0
        PC[pd.isnull(PC)] = 0
    C = SC.join(PC, how="outer").fillna(0)
    cosscores = 1-pairwise_distances(C[signatures.columns].T,C[phenotype.columns].T, metric="cosine")
    baseline_df = pd.DataFrame(cosscores, columns=["Cosine Score"], index=signatures.columns)
    return baseline_df

#######################################
## Compute accuracy of the predictor ##
#######################################
def empirical_pvalue(sorted_rewards, sorted_ground_truth, nperms, method="ks_2samp"):
    '''
        Compute an empirical p-value corresponding to testing whether the distributions in the predicted scores and the ground truth are similar
        by randomly permuting the values of the predictions and averaging the number of significant statistics across all permutations
        @param\tsorted_rewards\tPython float list: predicted scores sorted by genes of increasing predicted value
        @param\tsorted_ground_truth\tPython float list: ground truth scores sorted by genes of increasing predicted value
        @param\tnperms\tPython integer: number of permutations to perform
        @param\tmethod\tPython character string[default="ks_2samp"]: statistical test to perform at each permutation (should belong to scipy.stats)
        @return\tpvalue\tPython float: empirical p-value corresponding to the test across all @nperms permutations
    '''
    assert method in ["ks_2samp", "spearmanr", "kendalltau"]
    val_true, _ = eval(method)(sorted_rewards, sorted_ground_truth)
    p_cumsum, N = 0, len(sorted_rewards)
    for nperm in range(nperms):
        rseed(nperm)
        ## Sample a random ordering of elements
        ids__ = np.random.choice(range(N), size=N, replace=False, p=[1/N]*N)
        new_rewards = [sorted_rewards[i] for i in ids__]
        val_new, _ = eval(method)(new_rewards, sorted_ground_truth)
        p_cumsum += int(val_new >= val_true)
    return p_cumsum/nperms

def compute_metrics(rewards, ground_truth, K=[2,5,10], use_negative_class=False, nperms=10000, thres=0., beta=1.):
    '''
        Compute AUC, Hit Ratio @ k of method for positive/negative class with p-value
        @param\trewards\tPython float list: predicted scores
        @param\tground_truth\tPython float list: ground truth scores
        @param\tK\tPython integer list[default=[2,5,10]]: ranks at which the hit ratio should be computed
        @param\tuse_negative_class\tPython bool[default=False]: if set to True, compute the performance with respect to the negative class instead of the positive class
        @param\tnperms\tPython integer[default=10000]: number of permutations to perform
        @param\tthres\tPython float[default=0.]: decision threshold to determine the positive (resp. negative) class
        @param\tbeta\tPython float[default=1.]: value of the coefficient in the F-measure
        @return\tres_di\tPython dictionary: (keys=metrics, values=values of the metrics)
    '''
    ids = np.argsort(rewards).tolist()
    ids.reverse()
    sorted_rewards = [rewards[i] for i in ids]
    sorted_decisions = [int(rewards[i]>thres) for i in ids]
    sorted_ground_truth = [int(ground_truth[i]<0) if (use_negative_class) else int(ground_truth[i]>0) for i in ids]
    hr = [ACC(sorted_ground_truth[:k], sorted_decisions[:k]) for k in K]
    auc = AUC(sorted_ground_truth, sorted_rewards)
    acc = ACC(sorted_ground_truth, sorted_decisions)
    f_score = Fbeta(sorted_ground_truth, sorted_decisions, average='weighted', beta=beta)
    p = empirical_pvalue(sorted_rewards, sorted_ground_truth, nperms)
    res_di = {'AUC': np.round(auc,3), "ACC": np.round(acc,3), "p": np.round(p,3), ("F_%s") % beta: f_score}
    res_di.update({'HR@%d' % k : hr[ik] for ik, k in enumerate(K)})
    return res_di

######################
## DRUG SIMULATOR   ##
######################

def simulate(network_fname, targets, patients, score, simu_params={}, nbseed=0, quiet=False):
    '''
        Simulate and score the individual effects of drugs on patient phenotypes, compared to controls
        @param\tnetwork_fname\tPython character string: (relative) path to a network .BNET file
        @param\ttargets\tPandas DataFrame: rows/[genes] x columns/[drugs to test] (either 1: active expression, -1: inactive expression, 0: undetermined expression)
        @param\tpatients\tPandas DataFrame: rows/[genes] x columns/[samples] (either 1: activatory, -1: inhibitory, 0: no regulation).
        @param\tscore\tPython object: scoring of attractors
        @param\tsimu_params\tPython dictionary[default={}]: arguments to MPBN-SIM
        @param\tnbseed\tPython integer[default=0]
        @param\tquiet\tPython bool[default=False]
        @return\tscores\tPandas DataFrame: rows/[patient phenotypes] x columns/[drug names], values are the associated scores
    '''
    from multiprocessing import cpu_count
    assert simu_params.get('thread_count', 1)>=1 and simu_params.get('thread_count', 1)<=max(1,cpu_count()-2)
    ## Get M30 genes
    with open(network_fname, "r") as f:
        network = str(f.read())
    genes = [x.split(", ")[0] for x in network.split("\n") if (len(x)>0)]
    from random import seed as rseed
    rseed(nbseed)
    np.random.seed(nbseed)
    ## 1. Classification model b/w healthy and patient phenotypes (to classify final attractor states from treated patients)
    ## 2. Compute one score per drug and per patients
    if (simu_params.get('thread_count', 1)==1):
        scores = [simulate_treatment(network_fname, targets.loc[[g for g in targets.index if (g in genes)]], score, patients[[Patient]], simu_params, quiet=quiet) for Patient in patients.columns]
    else:
        scores = Parallel(n_jobs=simu_params['thread_count'], backend='loky')(delayed(simulate_treatment)(network_fname, targets.loc[[g for g in targets.index if (g in genes)]], score, patients[[Patient]], simu_params, quiet=quiet) for Patient in patients.columns)
    scores = pd.DataFrame(scores, index=patients.columns, columns=targets.columns)
    return scores

def compute_frontier(df, samples, nbseed=0, quiet=False):
    '''
        Fit a model to classify control/treated phenotypes
        @param\tdf\tPandas DataFrame: rows/[genes] x columns/[samples] (either 1: active expression, -1: inactive expression, 0: undetermined expression)
        @param\tsamples\tPandas DataFrame: rows/["annotation"] x columns/[samples ], values are 1 (healthy sample) or 2 (patient sample).
        @param\tnbseed\tPython integer[default=0]
        @param\tquiet\tPython bool[default=False]
        @return\tmodel\tPython object with a function "predict" that returns predictions (1: control, or 2: treated) on phenotypes
    '''
    from sklearn import svm
    model = svm.SVC(random_state=nbseed)
    X = df.values.T
    y = samples.values.T
    model.fit(X, y)
    acc = np.mean([int(x==y) for x,y in zip(list(model.predict(X)), list(y.T))])
    if (not quiet):
        print("<NORD_DS> Accuracy of the model %.2f" % acc)
    return model

def compute_score(f, x0, A, score, genes, nb_sims, experiments, repeat=1, exp_name="", quiet=False):
    '''
        Compute similarities between any attractor in WT and in mutants, weighted by their probabilities
        @param\tf\tBoolean Network (MPBN) object: the mutated network
        @param\tx0\tMPBN object: initial state
        @param\tA\tAttractor list: list of attractors in mutant network
        @param\tscore\tPython object: scoring of attractors
        @param\tgenes\tPython character string list: list of genes in the model @frontier
        @param\tnb_sims\tPython integer: number of iterations to compute the probabilities
        @param\texperiments\tPython dictionary list: list of experiments (different rates/depths)
        @param\trepeat\tPython integer[default=1]: how many times should these experiments be repeated
        @param\texp_name\tPython character string[default=""]: printable for an experiment
        @param\tquiet\tPython bool[default=False]
        @return\tscore\tPython float: change in attractors induced by the mutation
    '''
    for exp in experiments:
        if "name" not in exp:
            continue
        d_scores = []
        for _ in range(repeat):
            rates = getattr(mpbn_sim, f"{exp['rates']}_rates")
            depth = getattr(mpbn_sim, f"{exp['depth']}_depth")
            rates_args = exp.get("rate_args", {})
            depth_args = exp.get("depth_args", {})
            if (not quiet):
                print(exp_name+" "*int(len(exp_name)>0)+(f"- {depth.__name__}{depth_args}\t{rates.__name__}{rates_args}"))
            probs = mpbn_sim.estimate_reachable_attractors_probabilities(f, x0, A, nb_sims, depth(f, **depth_args), rates(f, **rates_args))
            attrs = pd.DataFrame({"MUT_%d"%ia: a for ia, a in enumerate(A)}).replace("*",np.nan).astype(float).loc[genes]
            probs = {i: x for i,x in list(probs.items()) if (x>0)}
            attrs = attrs[[attrs.columns[i] for i in list(probs.keys())]]
            ## if too many attractors, select the most common ones
            if (attrs.shape[1]>45000):
                idx_common = np.argsort([probs[i] for i in range(len(probs))]).tolist()
                idx_common.reverse()
                idx_common = idx_common[:45000]
                attrs = attrs[[attrs.columns[i] for i in idx_common]]
                probs = {i:probs[i] for i in idx_common}
                sprobs = sum(list(probs.values()))
                probs = {i:probs[i]*100/sprobs for i in probs}
            probs = np.array([probs[ia]/100 for ia in probs])
            classification_attrs = score(attrs)
            drug_score = probs.T.dot(classification_attrs)
            d_scores.append(drug_score)
    return np.mean(d_scores) if (repeat>1) else d_scores[0]

def simulate_treatment(network_name, targets, score, state, simu_params={}, quiet=False):
    '''
        Compute the score assigned to a drug with targets in @targets[[drug]] in network
        @param\tnetwork_name\tPython character string: filename of the network in .bnet (needs to be pickable)
        @param\ttargets\tPandas DataFrame: rows/[genes] x columns/[columns]
        @param\tscore\tPython object: scoring of attractors
        @param\tstate\tPandas DataFrame: binary patient initial state rows/[genes] x columns/[values in {-1,0,1}]
        @param\tsimu_params\tPython dictionary[default={}]: arguments to MPBN-SIM
        @param\tseednb\tPython integer[default=0]
        @param\tquiet\tPython bool[default=False]
        @return\teffects\tPython float dictionary: distance from attractors from treated networks to control profiles
    '''
    ## 1. Load the Boolean network
    f = mpbn.load(network_name)
    genes = list(state.index)
    ## 2. Create the initial profile
    if (not quiet):
        print("\t<NORD_DS> Initial state %s" % state.columns[0])
    x0 = f.zero()
    for i in list(state.loc[state[state.columns[0]]==1].index):
        x0[i] = 1
    ## 3. Create the mutated networks
    def patch_model(f, patch):
        f = mpbn.MPBooleanNetwork(f)
        for i, fi in patch.items():
            f[i] = fi
        return f
    experiments, nb_sims = [{"name": "mpsim", "rates": simu_params.get("rates", "fully_asynchronous"), "depth": simu_params.get("depth", "constant_unitary")}], simu_params.get("nb_sims", 100)
    mutants = {t: {gx: str(int(targets.loc[gx][t]>0)) for gx in targets[[t]].index if (targets.loc[gx][t]!=0)} for t in targets.columns}
    f_mutants = {name: patch_model(f, patch) for name, patch in mutants.items()}
    ## 4. Get the reachable attractors from initial state in the presence of mutations ("effect during drug exposure")
    ## 5. Estimate probabilities of attractors from Mutants and compute score
    effects = [0 if (t not in f_mutants) else compute_score(f_mutants[t], x0, [a for a in tqdm(list(f_mutants[t].attractors(reachable_from=x0)))], score, genes, nb_sims, experiments, exp_name="Drug %s (%d/%d) in state %s" % (t, it+1, targets.shape[1], state.columns[0]), quiet=quiet) for it, t in enumerate(targets.columns)]
    assert len(effects)==targets.shape[1]
    return effects
